
import numpy as np
from ...data import Dataset, Deviced, Batchable, Image_Dataset


@Dataset('unpaired-translation')
class UnpairedTranslationDataset(Image_Dataset):
	def __init__(self, A, dataset1=None, dataset2=None, sel_one=None, swap=None, **kwargs):

		if dataset1 is None:
			dataset1 = A.pull('dataset1')
		if dataset2 is None:
			dataset2 = A.pull('dataset2')

		if sel_one is None:
			sel_one = A.pull('sel-first', True)

		if swap is None:
			swap = A.pull('swap', False)

		if swap:
			dataset1, dataset2 = dataset2, dataset1

		print(f'Sizes: {len(dataset1)} vs {len(dataset2)}')

		super().__init__(A, din=dataset1.din, dout=dataset2.din, **kwargs)

		self.din1, self.dout1 = dataset1.din, dataset1.dout
		self.din2, self.dout2 = dataset2.din, dataset2.dout

		self.dataset1 = dataset1
		self.dataset2 = dataset2

		self.select_first = sel_one

	def __len__(self):
		return len(self.dataset1) * len(self.dataset2)

	def __getitem__(self, item):

		idx1 = item // len(self.dataset2)
		idx2 = item % len(self.dataset2)

		b1 = self.dataset1[idx1]
		b2 = self.dataset2[idx2]

		if self.select_first and isinstance(b1, (list, tuple)):
			b1 = b1[0]
			b2 = b2[0]

		return (b1, b2)


@Dataset('batched-unpaired-translation')
class BatchedUnpairedTranslationDataset(Batchable, UnpairedTranslationDataset):

	def __getitem__(self, item):

		if isinstance(item, (list, tuple)):
			item = np.array(item)

		idx1 = item // len(self.dataset2)
		idx2 = item % len(self.dataset2)

		b1 = self.dataset1[idx1]
		b2 = self.dataset2[idx2]

		if self.select_first and isinstance(b1, (list, tuple)):
			b1 = b1[0]
			b2 = b2[0]

		return (b1, b2)


# @Dataset('unpaired-translation')
# class UnpairedTranslationDataset(Batchable):
#
# 	def __init__(self, A, dataset1=None, dataset2=None, sel_one=None):
#
# 		mode = A.pull('mode', 'train')
#
# 		if dataset1 is None:
# 			dataset1 = A.pull('dataset1')
# 			if type(dataset1) == dict:
# 				slice1 = dataset1
# 				dataset1 = dataset1[mode]
# 			else:
# 				slice1 = {mode: dataset1}
# 		else:
# 			slice1 = None
# 		if dataset2 is None:
# 			dataset2 = A.pull('dataset2')
# 			if type(dataset2) == dict:
# 				slice2 = dataset2
# 				dataset2 = dataset2[mode]
# 			else:
# 				slice2 = {mode: dataset2}
# 		else:
# 			slice2 = None
#
# 		if sel_one is None:
# 			sel_one = A.pull('sel-first', True)
#
# 		super().__init__(dataset1.din, dataset2.din)
#
# 		self.mode = mode
# 		self._slice1, self._slice2 = slice1, slice2
#
# 		self.din1, self.dout1 = dataset1.din, dataset1.dout
# 		self.din2, self.dout2 = dataset2.din, dataset2.dout
#
# 		self.dataset1 = dataset1
# 		self.dataset2 = dataset2
#
# 		self.select_first = sel_one
#
# 	def __len__(self):
# 		return len(self.dataset1) * len(self.dataset2)
#
# 	def __getitem__(self, item):
#
# 		idx1 = item // len(self.dataset2)
# 		idx2 = item % len(self.dataset2)
#
# 		b1 = self.dataset1[idx1]
# 		b2 = self.dataset2[idx2]
#
# 		if self.select_first and isinstance(b1, (list, tuple)):
# 			b1 = b1[0]
# 			b2 = b2[0]
#
# 		return (b1, b2)
#
# 	def split(self, A):
#
# 		splits = {}
#
# 		cls = self.__class__
#
# 		for key in self._slice1.keys():
# 			if key == self.mode:
# 				splits[key] = self
# 			elif self._slice1[key] is not None and self._slice2[key] is not None:
# 				A.push('mode', key, silent=True)
# 				splits[key] = cls(A, dataset1=self._slice1[key], dataset2=self._slice2[key])
#
# 		return splits




import sys, os
import numpy as np
import torch
try:
	import matplotlib.pyplot as plt
	from matplotlib import animation
	from matplotlib.path import Path
	import matplotlib.patches as patches
	from matplotlib.figure import figaspect
except ImportError:
	print('WARNING: matplotlib not found')
try:
	import imageio
except ImportError:
	print('WARNING: imageio not found')
# try:
# 	import ffmpeg
# except ImportError:
# 	print('WARNING: ffmpeg not found')
try:
	from IPython.display import HTML
except ImportError:
	print('WARNING: ipython not found')

try:
	import pandas as pd
except ImportError:
	print('WARNING: pandas not found')
try:
	import seaborn as sns
except ImportError:
	print('WARNING: seaborn not found')

from torch.nn import functional as F

from .math import factors
from .data import load_images

def calc_tiling(N, H=None, W=None, prefer_tall=False):

	if H is not None and W is None:
		W = N//H
	if W is not None and H is None:
		H = N//W

	if H is not None and W is not None and N == H*W:
		return H, W

	H,W = tuple(factors(N))[-2:] # most middle 2 factors

	if H > W or prefer_tall:
		H, W = W, H

	# if not prefer_tall:
	# 	H,W = W,H
	return H, W


def tile_imgs(imgs, dim=0, H=None, W=None): # for numpy images

	raise NotImplementedError

	assert len(imgs.shape) > 3, 'not enough dimensions'
	if dim != 0:
		raise NotImplementedError

	N, h, w, c = imgs.shape

	if H is None or W is None:

		if H is None:
			W = 0

		pass

	assert H*W == imgs.shape[dim], 'Invalid tiling'


def plot_vec_fn(f, din=None, dim=None, lim=(-5, 5), N=1000, figax=None):
	if din is None:
		din = f.din
	if dim is not None:
		raise NotImplementedError
	
	if figax is None:
		figax = plt.subplots()
	fig, ax = figax
	plt.sca(ax)
	
	vals = torch.linspace(*lim, N)
	with torch.no_grad():
		outs = f(vals.unsqueeze(-1).expand(-1, din))
	plt.plot(vals, outs)
	plt.xlabel('Input = [1,...,1] * x')
	plt.ylabel('Output')
	return fig, ax

def plot_distribs(pts, figax=None, figsize=(6, 6), lim_y=None,
				  scale='count', inner='box', gridsize=100, cut=None, split=False,
				  color=None, palette=None, hue=None, **kwargs):
	
	if not isinstance(pts, list):
		pts = pts.tolist()
	
	inds = []
	vals = []
	for i, samples in enumerate(pts):
		inds.extend([i+1]*len(samples))
		vals.extend(samples)
	
	df = pd.DataFrame({'x': inds, 'y': vals})
	
	if figax is None:
		figax = plt.subplots(figsize=figsize)
	fig, ax = figax
	
	plt.sca(ax)
	
	# hue = None
	# split = False
	# color = 'C0'
	# inner = 'box'
	# palette = None
	
	if cut is not None:
		kwargs['cut'] = cut
	
	sns.violinplot(x='x', y='y', hue=hue,
				   data=df, split=split, color=color, palette=palette,
				   scale=scale, inner=inner, gridsize=gridsize, **kwargs)
	if lim_y is not None:
		plt.ylim(-lim_y, lim_y)
	# plt.title('Distributions of Latent Dimensions')
	# plt.xlabel('Dimension')
	# plt.ylabel('Values')
	# plt.tight_layout()
	# border, between = 0.02, 0.01
	# 	plt.subplots_adjust(wspace=between, hspace=between,
	# 						left=border, right=1 - border, bottom=border, top=1 - border)
	return fig, ax

def plot_parallel_coords(samples, categories=None, dim_names=None,
						 cat_styles=None, cat_names=None, include_legend=True,
						 mins=None, maxs=None,
						 figax=None, figsize=(8, 5), **default_style):
	'''
	samples: (N,D)
	categories: (N,)
	
	Example:

	from sklearn import datasets
	iris = datasets.load_iris()
	plot_parallel_coords(iris.data, dim_names=iris.feature_names,
						 categories=[iris.target_names[i] for i in iris.target])

	'''
	N, D = samples.shape
	
	classes = None
	if categories is not None:
		assert len(categories) == N, f'{len(categories)} vs {N}'
		
		classes = set(categories)
		if cat_names is not None:
			classes = cat_names
		K = len(classes)
		sep_classes = False
		if cat_styles is None:
			sep_classes = True
			cat_styles = {c: default_style.copy() for c in classes}
		
		for i, (name, style) in enumerate(cat_styles.items()):
			if 'color' in style and 'edgecolor' not in style:
				style['edgecolor'] = style['color']
				del style['color']
			elif sep_classes and 'edgecolor' not in style:
				style['edgecolor'] = f'C{i}'
			if 'facecolor' not in style:
				style['facecolor'] = 'none'
	
	try:
		samples = samples.cpu().numpy()
	except:
		pass
	
	if dim_names is None:
		dim_names = ['{}'.format(i) for i in range(D)]
	
	ynames = dim_names
	ys = samples
	
	if mins is None:
		mins = ys.min(axis=0)
	else:
		mins = mins.cpu().numpy()
	ymins = mins
	if maxs is None:
		maxs = ys.max(axis=0)
	else:
		maxs = maxs.cpu().numpy()
	ymaxs = maxs
	dys = ymaxs - ymins
	ymins -= dys * 0.05  # add 5% padding below and above
	ymaxs += dys * 0.05
	
	#     ymaxs[1], ymins[1] = ymins[1], ymaxs[1]  # reverse axis 1 to have less crossings
	dys = ymaxs - ymins
	
	# transform all data to be compatible with the main axis
	zs = np.zeros_like(ys)
	zs[:, 0] = ys[:, 0]
	zs[:, 1:] = (ys[:, 1:] - ymins[1:]) / dys[1:] * dys[0] + ymins[0]
	
	if figax is None:
		figax = plt.subplots(figsize=figsize)
	fig, host = figax
	
	axes = [host] + [host.twinx() for i in range(ys.shape[1] - 1)]
	for i, ax in enumerate(axes):
		ax.set_ylim(ymins[i], ymaxs[i])
		ax.spines['top'].set_visible(False)
		ax.spines['bottom'].set_visible(False)
		if ax != host:
			ax.spines['left'].set_visible(False)
			ax.yaxis.set_ticks_position('right')
			ax.spines["right"].set_position(("axes", i / (ys.shape[1] - 1)))
	
	host.set_xlim(0, ys.shape[1] - 1)
	host.set_xticks(range(ys.shape[1]))
	host.set_xticklabels(ynames, fontsize=14)
	host.tick_params(axis='x', which='major', pad=7)
	host.spines['right'].set_visible(False)
	host.xaxis.tick_top()
	#     host.set_title('Parallel Coordinates Plot — Iris', fontsize=18, pad=12)
	
	#     colors = plt.cm.Set2.colors
	legend_handles = {}
	for j in range(ys.shape[0]):
		# create bezier curves
		verts = list(zip([x for x in np.linspace(0, len(ys) - 1, len(ys) * 3 - 2, endpoint=True)],
						 np.repeat(zs[j, :], 3)[1:-1]))
		codes = [Path.MOVETO] + [Path.CURVE4 for _ in range(len(verts) - 1)]
		path = Path(verts, codes)
		
		if categories is None:
			style = default_style
		else:
			cls = categories[j]
			if cat_names is not None:
				cls = cat_names[cls]
			style = cat_styles[cls]
		
		patch = patches.PathPatch(path, **style)  # facecolor='none', lw=2, alpha=0.7, edgecolor=colors[iris.target[j]])
		
		host.add_patch(patch)
		
		if categories is not None:
			legend_handles[cls] = patch
	
	if include_legend:
		host.legend([legend_handles[c] for c in classes], classes,
					loc='lower center', bbox_to_anchor=(0.5, -0.18),
					ncol=len(classes), fancybox=True, shadow=True)
	
	return fig, host
	

def show_imgs(imgs, titles=None, H=None, W=None,
			  figsize=None, scale=1,
			  reverse_rows=False, grdlines=False,
			  channel_first=None,
			  imgroot=None, params={},
			  savepath=None, dpi=96, autoclose=True, savescale=1,
			  adjust={}, border=0., between=0.01):

	if isinstance(imgs, str):
		imgs = [imgs]
	if isinstance(imgs, (tuple, list)) and isinstance(imgs[0], str):
		imgs = list(load_images(*imgs, root=imgroot, channel_first=False))
		channel_first = False

	if isinstance(imgs, torch.Tensor):
		imgs = imgs.detach().cpu().squeeze(0).numpy()
	elif isinstance(imgs, (list, tuple)):
		imgs = [img.detach().cpu().numpy() if isinstance(img, torch.Tensor) else img for img in imgs]

	if isinstance(imgs, np.ndarray):
		shape = imgs.shape
		
		if channel_first is None \
				and shape[0 if len(shape) == 3 else 1] in {1, 3, 4} and shape[-1] not in {1, 3, 4}:
			channel_first = True
		
		# print(channel_first, shape)
		
		if len(shape) == 2 or (len(shape) == 3 and ((shape[0] in {1,3,4} and channel_first)
								  or (shape[-1] in {1,3,4} and not channel_first))):
			imgs = [imgs]
		elif len(shape) == 4:
			if channel_first:
				imgs = imgs.transpose(0,2,3,1)
				channel_first = False
		else:
			raise Exception(f'unknown shape: {shape}')

	imgs = [img.transpose(1,2,0).squeeze() if channel_first and len(img.shape)==3 else img.squeeze() for img in imgs]
	
	iH, iW = imgs[0].shape[:2]
	
	H,W = calc_tiling(len(imgs), H=H, W=W)
	
	fH, fW = iH*H, iW*W
	
	aw = None
	if figsize is None:
		aw, ah = figaspect(fH / fW)
		aw, ah = scale * aw, scale * ah
		figsize = aw, ah

	fg, axes = plt.subplots(H, W, figsize=figsize)
	axes = [axes] if len(imgs) == 1 else axes.flat

	hastitles = titles is not None and len(titles)
	if titles is None:
		titles = []
	if len(titles) != len(imgs):
		titles = titles + ([None] * (len(imgs) - len(titles)))

	for ax, img, title in zip(axes, imgs, titles):
		plt.sca(ax)
		if reverse_rows:
			img = img[::-1]
		plt.imshow(img, **params)
		if grdlines: # TODO: automate to allow more fine grain gridlines
			plt.plot([0, iW], [iH / 2, iH / 2], c='r', lw=.5, ls='--')
			plt.plot([iW / 2, iW / 2], [0, iH], c='r', lw=.5, ls='--')
			plt.xlim(0, iW)
			plt.ylim(0, iH)
			
		plt.axis('off')
		if title is not None:
			plt.title(title)
			
		# if title is not None:
		# 	plt.xticks([])
		# 	plt.yticks([])
		# 	plt.title(title)
		# else:
		# 	plt.axis('off')

	if hastitles and not len(adjust):
		plt.tight_layout()
	elif adjust is not None:
		
		base = dict(wspace=between, hspace=between,
							left=border, right=1 - border, bottom=border, top=1 - border)
		base.update(adjust)
		plt.subplots_adjust(**base)
	
	if savepath is not None:
		plt.savefig(savepath, dpi=savescale*(dpi if aw is None else fW/aw))
		if autoclose:
			plt.close()
			return
	
	return fg, axes


def plot_mat(M, val_fmt=None, figax=None, text_kwargs=dict(), **kwargs):
	if figax is None:
		figax = plt.subplots()
	fg, ax = figax
	plt.sca(ax)
	if isinstance(M, torch.Tensor):
		M = M.cpu().detach().numpy()
	if len(M.shape) == 1:
		M = M.reshape(1,-1)
	plt.matshow(M, False, **kwargs)
	eps = 0.03
	plt.subplots_adjust(eps,eps,1-eps,1-eps)
	if val_fmt is not None:
		if isinstance(val_fmt, int):
			val_fmt = f'.{val_fmt}f'
		if isinstance(val_fmt, str):
			val_fmt = '{:' + val_fmt + '}'
			fmt = lambda x: val_fmt.format(x)
		else:
			fmt = val_fmt
			
		if 'va' not in text_kwargs:
			text_kwargs['va'] = 'center'
		if 'ha' not in text_kwargs:
			text_kwargs['ha'] = 'center'
		for (i, j), z in np.ndenumerate(M):
			ax.text(j, i, fmt(z), **text_kwargs)


def play_back(imgs, figax=None, batch_first=True): # imgs is numpy: either (seq, H, W, C) or (batch, seq, H, W, C)
	if len(imgs.shape) > 4:  # tile first
		if not batch_first:
			imgs = imgs.transpose(1,0,2,3,4)
		N, S, h, w, c = imgs.shape
		H, W = calc_tiling(N)
		imgs = imgs.reshape(H, W, S, h, w, c)
		imgs = imgs.transpose(2, 0, 3, 1, 4, 5)
		imgs = imgs.reshape(S, H * h, W * w, c)

	if figax is None:
		figax = plt.subplots()
	fig, ax = figax
	plt.sca(ax)
	for i, img in enumerate(imgs):
		plt.cla()
		plt.axis('off')
		plt.imshow(img)
		plt.title('{}/{}'.format(i + 1, len(imgs)))
		plt.tight_layout()
		plt.pause(0.02)
	return fig, ax

def plot_stat(all_stats, key, xdata=None, figax=None, **fmt_args): # all_stats should be a list of StatsMeters containing key
	if figax is None:
		figax = plt.subplots()
	fig, ax = figax

	if xdata is not None:
		assert len(all_stats) == len(xdata)
	else:
		xdata = np.arange(1, len(all_stats)+1)

	ydata = np.array([stat[key].avg.item() for stat in all_stats])

	plt.sca(ax)
	plt.plot(xdata, ydata, **fmt_args)

	return figax

def fig_to_rgba(fig):
	fig.canvas.draw()
	w, h = fig.canvas.get_width_height()
	buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
	buf.shape = (h, w, 4)
	buf = np.roll(buf, 3, axis=2)
	return buf

def flatten_tree(tree, is_tree=None, prefix=None): # returns tuples of deep keys
	if prefix is None:
		prefix = []

	flat = []

	for k, v in tree.items():
		prefix.append(k)
		if (is_tree is not None and is_tree(v)) or isinstance(v, dict):
			new = flatten_tree(v, is_tree=is_tree, prefix=prefix)
			if len(new):
				flat.extend(new)
				prefix.pop()
				continue
		flat.append(tuple(prefix))
		prefix.pop()

	return flat

class Video(object):
	def __init__(self, frames): # frames must be a sequence of numpy byte images
		self.frames = frames
		# self.path = path

	def play(self, mode='mpl', scale=1):

		if mode == 'mpl':
			play_back(self.frames)
		elif mode == 'jupyter':
			return HTML(self.as_animation(scale=scale).to_html5_video())
		else:
			raise Exception('Unknonwn mode: {}'.format(mode))

	def as_animation(self, scale=1, fps=20):

		H, W, C = self.frames[0].shape

		asp = W/H

		fig = plt.figure(figsize=(asp, 1), dpi=int(H*scale),)

		ax = plt.axes([0, 0, 1, 1], frameon=False)
		ax.get_xaxis().set_visible(False)
		ax.get_yaxis().set_visible(False)
		plt.autoscale(tight=True)

		im = plt.imshow(self.frames[0])
		# plt.axis('off')
		# plt.tight_layout()

		plt.close()

		def init():
			im.set_data(self.frames[0])

		def animate(i):
			im.set_data(self.frames[i])
			return im

		anim = animation.FuncAnimation(fig, animate, init_func=init, frames=len(self.frames), interval=1000//fps)

		return anim

	def export(self, path, fmt='mp4'):

		assert fmt in {'mp4', 'gif', 'jpg', 'png'}

		if fmt in {'mp4', 'gif'}:
			imageio.mimsave(path, self.frames)

		elif fmt in {'jpg', 'png'}:
			raise NotImplementedError

		# elif fmt == 'mp4':
		# 	imageio.mimwrite('test2.mp4', self.frames, fps=30)

			# vcodec = 'mp4'
			# framerate = 25
			#
			# images = self.frames
			# n, height, width, channels = images.shape
			# process = (
			# 	ffmpeg
			# 		.input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(width, height))
			# 		.output(path, pix_fmt='yuv420p', vcodec=vcodec, r=framerate)
			# 		.overwrite_output()
			# 		.run_async(pipe_stdin=True)
			# )
			# for frame in images:
			# 	process.stdin.write(
			# 		frame
			# 			.astype(np.uint8)
			# 			.tobytes()
			# 	)
			# process.stdin.close()
			# process.wait()

class Animation(Video):

	def __init__(self, anim):
		super().__init__(None) # no frames
		self.anim = anim

	def play(self, mode='jupyter', scale=1):

		if mode != 'jupyter':
			raise NotImplementedError

		if scale != 1:
			print('WARNING: scale has no effect for animations that are already done')

		return HTML(self.anim.to_html5_video())

	def as_animation(self, scale=1, fps=None):
		if scale != 1 or fps is not None:
			print('WARNING: scale/fps has no effect for animations that are already done')
		return self.anim

	def export(self, path, fmt=None, fps=20):

		assert not os.path.isdir(path), 'path is already a dir'

		if fmt is None:
			fmt = os.path.basename(path).split('.')[-1]
			if fmt in {'png', 'jpg'}:
				fmt = 'frames'

		if fmt == 'mp4':
			Writer = animation.writers['ffmpeg']
			writer = Writer(fps=fps, metadata=dict(artist='Me'), bitrate=1800)
			self.anim.save(path, writer=writer)

		elif fmt == 'gif':
			self.anim.save(path, writer='imagemagick', fps=fps)

		elif fmt == 'frames':
			self.anim.save(path, writer="imagemagick")


class Pretty_Formatter(object):
	def __init__(self, lf='\n', ht='  '):
		self.types = {}
		self.lf, self.ht = lf, ht
		self.reset()

	def set_formater(self, typ, callback):
		self.types[typ] = callback

	def reset(self):
		self.types = {
			object: self.format_object,
			dict: self.format_dict,
			list: self.make_iter_format(['['], [']']),
			set: self.make_iter_format(['{'], ['}']),
			tuple: self.make_iter_format(['('], [')']),
		}

	def _indent(self, num=None):
		return self.lf + self.ht*num

	def __call__(self, value, formatters={}, lf=None, ht=None):

		self.types.update(formatters)

		if lf is not None:
			self.lf = lf
		if ht is not None:
			self.ht = ht

		deep_lines = self._format(value)

		s = self._flatten(deep_lines)
		if s[0] == self.lf:
			s = s[1:]
		return s

	def _format(self, value):
		out = None
		for cls in value.__class__.__mro__:
			if cls in self.types:
				out = self.types[cls](value, self._format) # should return a list of tuples/strings/lists
				break

		if out is None:
			print(f'WARNING: unable to format {value}')

		return out

	def _flatten(self, line, indent=0):

		if type(line) == str:
			return line.replace(self.lf, self._indent(indent+2))

		if type(line) == tuple:
			return ''.join(self._flatten(w, indent=indent) for w in line)

		if type(line) == list:
			ind = self._indent(indent+1)
			lines = [self._flatten((l,) if type(l) == str else l, indent=indent+1)
							for l in line]
			start = ind.join(lines[:-1])
			ret = self._indent(indent)
			start += ret + lines[-1]
			return start

		raise Exception(f'not recognized: {line}')


	@staticmethod
	def format_object(value, fmt):
		return repr(value)

	@staticmethod
	def format_dict(value, fmt):
		lines = []
		lines.append('{')
		lines.extend((repr(key),': ', fmt(val)) for key, val in value.items())
		# lines.extend((repr(key), ': ', fmt(val), ',') for key, val in value.items())
		lines.append('}')
		return lines

	@staticmethod
	def make_iter_format(start, end):
		def _iter_format(value, fmt):
			nonlocal start, end
			lines = []
			lines.extend(start)
			lines.extend((fmt(val)) for val in value)
			# lines.extend((fmt(val), ',') for val in value)
			lines.extend(end)
			return lines
		return _iter_format

pretty_format = Pretty_Formatter()


def image_size_limiter(imgs, size=(128,128)):
	H, W = imgs.shape[-2:]
	
	mH, mW = size
	
	if H <= mH and W <= mW:  # allows upto around 128x128
		return imgs
	
	imgs = F.interpolate(imgs, size=(128, 128))
	return imgs


###################
# Flow visualization - using yiq

_yiq_to_rgb = torch.FloatTensor([[1, 0.956, 0.621],
						[1, -0.272, -0.647],
						[1, -1.106, 1.703]])
_iq_to_rgb = torch.FloatTensor([[0.956, 0.621],
						[-0.272, -0.647],
						[-1.106, 1.703]])
_rgb_to_yiq = torch.FloatTensor([[.299, .587, 0.114],
						   [0.596, -0.274, -0.322],
						   [0.211, -0.523, 0.312]])

# for 2d flow
def iq2rgb(img, clamp=1): # reqs B x C=2 x H x W
	B, C, H, W = img.size()
	assert C == 2
	assert clamp > 0
	#print(img.permute(1,0,2,3).contiguous().view(2, -1).contiguous().clamp(-clamp, clamp).size())
	#quit()
	img = img.permute(1,0,2,3).contiguous().view(2, -1).contiguous().clamp(-clamp, clamp) * torch.Tensor([[0.5957], [0.5226]]).type_as(img) / clamp
	#print(img.abs().max(-1)[0])
	img = _iq_to_rgb.type_as(img) @ img + 0.5
	return img.view(3,B,H,W).permute(1,0,2,3).clamp(0,1)

# for 3d flow
def yiq2rgb(img, clamp=1): # reqs B x C=3 x H x W
	B, C, H, W = img.size()
	assert C == 3
	assert clamp > 0
	#print(img.permute(1,0,2,3).contiguous().view(2, -1).contiguous().clamp(-clamp, clamp).size())
	#quit()
	img = img.permute(1,0,2,3).contiguous().view(3, -1).contiguous().clamp(-clamp, clamp) * torch.Tensor([[0.5957], [0.5226], [1]]).type_as(img) / clamp
	#print(img.abs().max(-1)[0])
	img = _yiq_to_rgb.type_as(img) @ img + 0.5
	return img.view(3,B,H,W).permute(1,0,2,3).clamp(0,1)





# def tile_images(img_nhwc):
#     """
#     Tile N images into one big PxQ image
#     (P,Q) are chosen to be as close as possible, and if N
#     is square, then P=Q.
#
#     input: img_nhwc, list or array of images, ndim=4 once turned into array
#         n = batch index, h = height, w = width, c = channel
#     returns:
#         bigim_HWc, ndarray with ndim=3
#     """
#     img_nhwc = np.asarray(img_nhwc)
#     N, h, w, c = img_nhwc.shape
#     H = int(np.ceil(np.sqrt(N)))
#     W = int(np.ceil(float(N)/H))
#     img_nhwc = np.array(list(img_nhwc) + [img_nhwc[0]*0 for _ in range(N, H*W)])
#     img_HWhwc = img_nhwc.reshape(H, W, h, w, c)
#     img_HhWwc = img_HWhwc.transpose(0, 2, 1, 3, 4)
#     img_Hh_Ww_c = img_HhWwc.reshape(H*h, W*w, c)
#     return img_Hh_Ww_c





import numpy as np

import torch
from collections import deque, OrderedDict

import omnifig as fig

from .features import Configurable, Switchable

# class StatsCollector(object):
# 	def __init__(self, require_all_keys=True):
# 		assert require_all_keys
# 		self.bag = None
# 		self.steps = deque()
#
# 	def update(self, step, stats):
# 		if self.bag is None:
# 			self.bag = {stat:deque() for stat in stats}
# 		self.steps.append(step)
# 		for stat, val in stats.items():
# 			self.bag[stat].append(val)
#
# 	def export(self):
# 		data = {'steps':np.array(self.steps)}
# 		if self.bag is not None:
# 			data['stats'] = {stat:np.array(data) for stat, data in self.bag.items()}
# 		return data
#
# 	def save(self, path):
# 		torch.save(self.export(), path)


# def override_all_stats_tau(tau): # TODO: track all stats objs and then update tau
# 	pass




# @fig.AutoModifier('reg-stats')
# class RegStats(StatsContainer):
# 	def __init__(self, A, **kwargs):
#
# 		super().__init__(A, **kwargs)
#
# 		self._init_stats(A)
#
# 	def _init_stats(self, A=None):
# 		if A is not None:
# 			A.push('stats._type', 'stats', overwrite=False, silent=True)
# 			self.stats = A.pull('stats')
#
# 			if A.pull('_reg_stats', True):
# 				records = A.pull('records', None, ref=True)
# 				if records is not None:
# 					stats_fmt = A.pull('_stats_fmt', None)
# 					records.register_stats_client(self, fmt=stats_fmt)

@fig.Component('stats')
class StatsMeter(Configurable, OrderedDict):
	def __init__(self , A=None, meter_info=None, **kwargs):
		
		if meter_info is None:
			assert A is not None, 'need config to know how to create new meters'
			if 'meter-info' not in A:
				A.push('meter-info._type', 'meter', overwrite=False)
			meter_info = A.pull('meter-info', raw=True)
		
		super().__init__(A, **kwargs)
		
		self.meter_info = meter_info
		
	def set_tau(self, tau):
		for meter in self.values():
			meter.tau = tau

	def reset(self):
		for meter in self.values():
			meter.reset()
	
	# def __deepcopy__(self):
	# 	'''returns deepcopy of self'''
	# 	new = self.__class__(meter_info=self.meter_info)
	# 	for name, meter in self.items():
	# 		new[name] = meter.copy()
	# 	return new

	def _create_meter(self, name=None, tau=None):
		meter = self.meter_info.pull_self(silent=True)
		if tau is not None:
			meter.set_tau(tau)
		return meter

	def new(self, *names, tau=None):
		for name in names:
			if name not in self:
				self[name] = self._create_meter(name, tau=tau)

	def discard(self, *names):
		for name in names:
			if name in self:
				del self[name]
	def remove(self, *names):
		for name in names:
			del self[name]

	def mete(self, name, value, n=1):
		# assert isinstance(value, (int,float)) or value.size == 1, 'unknown: {} {}'.format(value.shape, value)
		self[name].mete(value, n=n)
	
	def __add__(self, other):
		new = self.copy()
		new.join(other)
		return new
	
	def join(self, other, intersection=False, fmt=None):
		'''other is another stats meter'''
		for name, meter in other.items():
			if fmt is not None:
				name = fmt.format(name)
			if name in self:
				self[name].join(meter)
			elif not intersection:
				self[name] = meter.copy()

	def shallow_join(self, other, fmt=None):
		for name, meter in other.items():
			if fmt is not None:
				name = fmt.format(name)
			self[name] = meter
	
	def vals(self, fmt='{}'):
		return {fmt.format(k):v.val.item() for k,v in self.items()}

	def avgs(self, fmt='{}'):
		return {fmt.format(k):v.avg.item() for k,v in self.items() if v.count > 0}
	
	def smooths(self, fmt='{}'):
		return {fmt.format(k): (v.smooth.item()) for k,v in self.items() if v.smooth is not None}

	def split(self):
		all_vals = {k:v.export() for k,v in self.__dict__.items()}
		key = next(iter(self.__dict__.keys()))
		try:
			shape = all_vals[key]['val'].shape
		except Exception as e:
			raise e

		stats = [{stat:{k:v[idx] for k,v in vals.items()} for stat,vals in all_vals.items()} for idx in np.ndindex(*shape)]

		return [StatsMeter(**info) for info in stats]

	def load(self, stats):
		self.update({stat:self._create_meter(stat).load(vals) for stat,vals in stats.items()})

	def export(self):
		return {name : meter.export() for name, meter in self.items()}
	#
	# def save(self, filename):
	# 	torch.save(self.export() ,filename)
	#
	# def load_dict(self, d):
	# 	for name, meter in d.items():
	# 		self._stats[name] = meter
	#
	# def load(self, filename):
	# 	self.load_dict(torch.load(filename))
	
	def __repr__(self):
		return self.__str__()
		
	def __str__(self):
		return 'StatsMeter({})'.format(', '.join(['{}:{:.4f}'.format(k,v.val.item()) for k,v in self.items()]))


@fig.Component('stats-manager')
class StatsManager(Switchable, StatsMeter):

	def __init__(self, A=None, collections=None, collection_fmts=None, **kwargs):
		
		if collection_fmts is None and A is not None:
			collection_fmts = A.pull('stat-collection-fmts', {})
		if collections is None:
			collections = {}
		
		reset_collections = set(A.pull('reset-modes', []))
		
		super().__init__(A=A, **kwargs)
		self._collections = collections
		self._reset_collections = reset_collections
		self._master_names = set()
		self._timestep = None
		self._archive = {}
		self._collection_fmts = collection_fmts
	
	def vals(self, fmt=None):
		if fmt is None:
			fmt = self._collection_fmts.get(self.get_mode(), '{}')
		return super().vals(fmt=fmt)
	
	def avgs(self, fmt=None):
		if fmt is None:
			fmt = self._collection_fmts.get(self.get_mode(), '{}')
		return super().avgs(fmt=fmt)
	
	def smooths(self, fmt=None):
		if fmt is None:
			fmt = self._collection_fmts.get(self.get_mode(), '{}')
		return super().smooths(fmt=fmt)
	
	def new(self, *names, tau=None):
		self._master_names.update(names)
		super().new(*names, tau=tau)
	
	def remove(self, *names):
		self._master_names.difference_update(names)
		super().remove(*names)

	def discard(self, *names):
		self._master_names.difference_update(names)
		super().discard(*names)
		
	def _create_collection(self):
		return {name:self._create_meter(name) for name in self._master_names}
	
	def set_step(self, ticks):
		self._timestep = ticks
	
	def archive(self, ticks=None, mode=None):
		if mode is None:
			mode = self.get_mode()
		if ticks is None:
			ticks = self._timestep
		if mode is None or ticks is None:
			return
		
		self._sync_collections()
		
		if mode not in self._archive:
			self._archive[mode] = {}
		self._archive[mode][str(ticks)] = {name: meter.export() for name, meter in self._collections[mode].items()}
	
	def switch_to(self, mode):
		
		old = self.get_mode()
		if mode == old:
			return
		
		self.archive()
		
		super().switch_to(mode)
		self.clear()
		
		if mode not in self._collections or mode in self._reset_collections:
			self._collections[mode] = self._create_collection()
		self.update(self._collections[mode])
		
	def _sync_collections(self):
		mode = self.get_mode()
		if mode is not None:
			self._collections[self.get_mode()] = dict(self)
		
	def export(self):
		self.archive()
		return self._archive
		
	def load(self, stats):
		mode = None
		self._archive = stats
		self._collections = {}
		for mode, arc in stats.items():
			last = arc[str(max(map(int,arc.keys())))]
			if mode not in self._reset_collections:
				self._collections[mode] = {name:self._create_meter(name).load(vals)
				                                for name, vals in last.items()}
		
		if mode is not None:
			self._master_names = set(last.keys())
		
		self.clear()
		self.mode = None


class StatsClient(Configurable):
	def __init__(self, A, **kwargs):
		stats = A.pull('stats', None, ref=True)
		if stats is None:
			print('WARNING: no stats manager found')
		fmt = A.pull('stats-fmt', None)
		
		default_tau = A.pull('smooth-tau', '<>tau', None)
		
		super().__init__(A, **kwargs)
		
		self._stats = stats
		self._stats_fmt = fmt
		self._stats_tau = default_tau
	
	def register_stats(self, *names):
		if self._stats is not None:
			if self._stats_fmt is not None:
				names = [self._stats_fmt.format(name) for name in names]
			self._stats.new(*names, tau=self._stats_tau)
	
	def mete(self, name, val, n=1):
		if self._stats is not None:
			if self._stats_fmt is not None:
				name = self._stats_fmt.format(name)
			self._stats.mete(name, val, n=n)
	
	def get_stat(self, name):
		if self._stats is not None:
			if self._stats_fmt is not None:
				name = self._stats_fmt.format(name)
			return self._stats.get(name, None)


_tau = 0.005
def set_default_tau(tau):
	global _tau
	_tau = tau

### Computes sum/avg stats
@fig.Component('meter')
class AverageMeter(Configurable):
	"""Computes and stores the average and current value"""
	def __init__(self, A=None, tau=None, **kwargs):
		super().__init__(A, **kwargs)
		
		self.reset()
		
		if tau is None:
			assert A is not None, 'no config provided'
			tau = A.pull('smooth-tau', '<>tau', _tau)
		self.set_tau(tau)
		
	def set_tau(self, tau):
		self.tau = torch.tensor(tau)
		
	def copy(self):
		new = self.__class__(tau=self.tau)
		for k, v in self.__dict__.items():
			if k == 'tau':
				new.tau = self.tau
				continue
			try:
				new.__dict__[k] = None if v is None else v.clone()
			except Exception as e:
				print(k, type(v), v)
				raise e
		return new
	
	def reset(self):
		self.val = torch.tensor(0.)
		self.avg = torch.tensor(0.)
		self.sum = torch.tensor(0.)
		self.count = torch.tensor(0.)
		self.max = None
		self.min = None
		self.smooth = None
		self.var = torch.tensor(0.)
		self.std = torch.tensor(0.)
		self.S = torch.tensor(0.)
	
	def mete(self, val, n=1):
		try:
			val = val.float().detach().cpu()
		except:
			val = torch.tensor(val).float()
		self.val = val
		self.sum += self.val * n
		prev_count = self.count
		self.count += n
		delta = self.val - self.avg
		self.S += delta**2 * n * prev_count / self.count
		self.avg = self.sum / self.count
		self.max = self.val if self.max is None else np.maximum(self.max, val)
		self.min = self.val if self.min is None else np.minimum(self.min, val)
		self.var = self.S / self.count
		self.std = torch.sqrt(self.var)
		self.smooth = self.val if self.smooth is None else (self.smooth*(1-self.tau) + self.val*self.tau)

	def __len__(self):
		return self.count.item()

	def load(self, vals):
		for k, v in vals.items():
			self.__dict__[k] = torch.tensor(v)
		return self
	
	def export(self): #
		if self.val.nelement() > 1:
			return {k: v.detach().cpu().numpy() for k, v in self.__dict__.items() if v is not None}
		else:
			return {k: v.item() for k,v in self.__dict__.items() if v is not None}

	def type(self):
		return '{}-{}'.format(self.val.type(), tuple(self.val.size()))
		
	def join(self, other):
		'''other is another average meter, assumed to be more uptodate (for val)
			this does not mutate other'''
		self.val = other.val
		if self.count+other.count == 0: # both are empty
			self.reset()
			return

		delta = other.avg - self.avg
		prev_count = self.count
		self.count += other.count
		self.S += other.S + delta ** 2 * other.count * prev_count / self.count
		self.var = self.S / self.count
		self.std = np.sqrt(self.var)

		self.sum += other.sum
		self.avg = self.sum / self.count

		if self.max is None:
			self.max = other.max
		elif other.max is not None:
			self.max = max(self.max, other.max)
		if self.min is None:
			self.min = other.min_val
		elif other.min_val is not None:
			self.min = min(self.min, other.min_val)




import torch
import torch.nn as nn
from torch.distributions import Normal as NormalDistribution
# from .. import framework as fm
from ..op import framework as fm

from omnibelt import InitWall, unspecified_argument
import omnifig as fig

from .layers import make_MLP

#################
# region General
#################

class MLPBase(fm.FunctionBase, InitWall, nn.Sequential):
	def __init__(self, din, dout, hidden=None, initializer=None,
	             nonlin='elu', output_nonlin=None,
	             bias=True, output_bias=None, make_kwargs={}, _req_args=(), **kwargs):

		assert len(_req_args) == 0, str(_req_args)

		net = make_MLP(din=din, dout=dout, hidden=hidden, initializer=initializer,
		               nonlin=nonlin, output_nonlin=output_nonlin,
		               bias=bias, output_bias=output_bias, **make_kwargs)
		super().__init__(din=net.din, dout=net.dout, _req_args=tuple(net), **kwargs)

@fig.Component('mlp')
class MLP(fm.Function, MLPBase):
	def __init__(self, A, din=unspecified_argument, dout=unspecified_argument,
	             hidden=unspecified_argument, initializer=unspecified_argument,
	             nonlin=unspecified_argument, output_nonlin=unspecified_argument,
	             bias=unspecified_argument, output_bias=unspecified_argument,
	             make_kwargs=unspecified_argument, **kwargs):

		# region fill in args

		if din is unspecified_argument:
			din = A.pull('_din', '<>din')
		if dout is unspecified_argument:
			dout = A.pull('_dout', '<>dout')
		if hidden is unspecified_argument:
			hidden = A.pull('hidden_dims', '<>hidden_fc', '<>hidden', None)
		if initializer is unspecified_argument:
			initializer = A.pull('initializer', None)
		if nonlin is unspecified_argument:
			nonlin = A.pull('nonlin', 'elu')
		if output_nonlin is unspecified_argument:
			output_nonlin = A.pull('output_nonlin', '<>out_nonlin', None)
		if bias is unspecified_argument:
			bias = A.pull('bias', True)
		if output_bias is unspecified_argument:
			output_bias = A.pull('output_bias', '<>out_bias', None)
		if make_kwargs is unspecified_argument:
			make_kwargs = A.pull('make_kwargs', {})

		# endregion

		super().__init__(A, din=din, dout=dout,
		                 hidden=hidden, initializer=initializer,
		                 nonlin=nonlin, output_nonlin=output_nonlin,
		                 bias=bias, output_bias=output_bias,
		                 make_kwargs=make_kwargs, **kwargs)

@fig.Component('multihead')
class Multihead(fm.FunctionBase): # currently, the input dim for each head must be the same (but output can be different) TODO: generalize
	def __init__(self, A):
		
		din = A.pull('din')
		pin, N = din
		
		n_first = A.pull('n_first', False) # to change the "pin" and "N" dimensions
		if n_first:
			raise NotImplementedError
		
		merge = A.pull('merge', 'concat')
		if merge != 'concat':
			raise NotImplementedError
		
		A.push('heads.din', pin, silent=True)
		create_head = A.pull('heads')
		
		head_douts = A.pull('head_douts', None)
		try:
			len(head_douts)
		except TypeError:
			idouts = iter([head_douts]*len(create_head))
		else:
			idouts = iter(head_douts)
		
		heads = []
		
		nxt = create_head.view()
		if head_douts is not None:
			nxt.push('dout', next(idouts), silent=True)
		for head in create_head:
			heads.append(head)
			try:
				nxt = create_head.view()
			except StopIteration:
				break
			else:
				if head_douts is not None:
					nxt.push('dout', next(idouts), silent=True)
		
		assert N == len(heads), f'{N} vs {len(heads)}'
		
		douts = [head.dout for head in heads]
		
		if merge == 'concat':
			dout = sum(douts)
		
		super().__init__(din, dout)
		
		self.heads = nn.ModuleList(heads)
		
		self.merge = merge
		
	def forward(self, xs):
		
		xs = xs.permute(2,0,1)
		
		ys = [head(x) for head, x in zip(self.heads, xs)]
		
		return torch.cat(ys, dim=1)

@fig.Component('multilayer') # used for CNNs
class MultiLayer(fm.Function):

	def __init__(self, A, layers=None, din=None, dout=None, in_order=None, **kwargs):

		if layers is None:
			layers = self._create_layers(A, din=din, dout=dout, in_order=in_order)

		super().__init__(A, din=layers[0].din, dout=layers[-1].dout, **kwargs)

		self.layers = nn.ModuleList(layers)

	def _create_layers(self, A, din=None, dout=None, in_order=None):

		if din is None:
			din = A.pull('_din', '<>din', None)
		if dout is None:
			dout = A.pull('_dout', '<>dout', None)

		assert din is not None or dout is not None, 'need some input info'

		if in_order is None:
			in_order = A.pull('in_order', din is not None)

		self._current = din if in_order else dout
		self._req_key = 'din' if in_order else 'dout'
		self._find_key = 'dout' if in_order else 'din'
		end = dout if in_order else din

		pre, post = ('first', 'last') if in_order else ('last', 'first')

		pre_layer_info = A.pull(pre, None, raw=True)
		pre_layer = None if pre_layer_info is None else self._create_layer(pre_layer_info)

		mid = []
		create_layers = A.pull('layers', None, as_iter=True)
		if create_layers is not None:
			create_layers.set_auto_pull(False)  # prevents layer from being created before din/dout info is updated
			create_layers.set_reversed(not in_order)
	
			mid = [self._create_layer(layer) for layer in create_layers]

		post_layer = None
		post_layer_info = A.pull(post, None, raw=True)
		if end is not None or post_layer_info is not None:
			if post_layer_info is None:
				A.push('output_nonlin', None)
				post_layer_info = A.sub(post)
			mytype = post_layer_info.push('_type', 'dense-layer', silent=True, overwrite=False)
			if end is not None:
				post_layer_info.push(self._find_key, end, silent=True)
			if mytype == 'dense-layer' and in_order:
				post_layer_info.push('nonlin', '<>output_nonlin', silent=False, overwrite=False)
			post_layer = self._create_layer(post_layer_info, empty_find=end is None)

		layers = []
		if pre_layer is not None:
			layers.append(pre_layer)
		layers.extend(mid)
		if post_layer is not None:
			layers.append(post_layer)
		return layers if in_order else layers[::-1]

	def _create_layer(self, info, empty_find=True):
		current = self._current
		info.push(self._req_key, current, silent=True, overwrite=True)
		if empty_find:
			info.push(self._find_key, None, silent=True, overwrite=True)
		layer = info.pull_self()
		# c, n = self._current, getattr(layer, self._find_key)
		self._current = getattr(layer, self._find_key)
		return layer

	def forward(self, x):
		for i, layer in enumerate(self.layers):
			x = layer(x)
		return x

# endregion
#################












import omnifig as fig

from .. import util


dataset_registry = util.Registry()

def register_dataset(name, dataset):
	cmpn_name = f'dataset/{name}'
	fig.Component(cmpn_name)(dataset)
	
	dataset_registry.new(name, cmpn_name)


def Dataset(name):
	def _reg_fn(fn):
		nonlocal name
		register_dataset(name, fn)
		fn._dataset_ident = name
		return fn
	
	return _reg_fn






import torch
import re
from torch._six import container_abcs, string_classes, int_classes

from ..util import TensorDict, TensorList

np_str_obj_array_pattern = re.compile(r'[SaUO]')

default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}")

def _collate_movable(batch):
	r"""Puts each data field into a tensor with outer dimension batch size"""

	elem = batch[0]
	elem_type = type(elem)
	if isinstance(elem, torch.Tensor):
		out = None
		if torch.utils.data.get_worker_info() is not None:
			# If we're in a background process, concatenate directly into a
			# shared memory tensor to avoid an extra copy
			numel = sum([x.numel() for x in batch])
			storage = elem.storage()._new_shared(numel)
			out = elem.new(storage)
		return torch.stack(batch, 0, out=out)
	elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
			and elem_type.__name__ != 'string_':
		elem = batch[0]
		if elem_type.__name__ == 'ndarray':
			# array of string classes and object
			if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
				raise TypeError(default_collate_err_msg_format.format(elem.dtype))

			return _collate_movable([torch.as_tensor(b) for b in batch])
		elif elem.shape == ():  # scalars
			return torch.as_tensor(batch)
	elif isinstance(elem, float):
		return torch.tensor(batch, dtype=torch.float64)
	elif isinstance(elem, int_classes):
		return torch.tensor(batch)
	elif isinstance(elem, string_classes):
		return batch
	elif isinstance(elem, container_abcs.Mapping):
		return TensorDict({key: _collate_movable([d[key] for d in batch]) for key in elem})
	elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
		return elem_type(*(_collate_movable(samples) for samples in zip(*batch)))
	elif isinstance(elem, container_abcs.Sequence):
		transposed = zip(*batch)
		return TensorList([_collate_movable(samples) for samples in transposed])

	raise TypeError(default_collate_err_msg_format.format(elem_type))




import torch
from torch import nn
from torch.distributions import Normal as NormalDistribution

import omnifig as fig

# import torch
# import torch.nn as nn
# from .. import framework as fm
# from ..op import framework as fm

# from omnibelt import InitWall, unspecified_argument
# import omnifig as fig

# from .layers import make_MLP

from ..op.framework import Function
from .nets import MultiLayer
from .. import util

###############
# region Prior
###############

class Prior(Function):

	def __init__(self, A, prior_dim=None, **kwargs):
		super().__init__(A, **kwargs)
		# if prior_dim is None:
		# 	prior_dim = A.pull('latent-dim', '<>din', None, silent=True)
		self.prior_dim = prior_dim

	def sample_prior(self, *shape):
		raise NotImplementedError

@fig.AutoModifier('gaussian-prior')
class Gaussian(Prior):
	def sample_prior(self, N=1, *D):
		if not len(D):
			D = self.prior_dim
			if isinstance(D, int):
				D = (D,)
		return torch.randn(N, *D, device=self.device)


@fig.AutoModifier('uniform-prior')
class Uniform(Prior):
	def sample_prior(self, N=1, *D):
		if not len(D):
			D = self._prior_dim
			if isinstance(D, int):
				D = (D,)
		return torch.rand(N, *D, device=self.device)


@fig.AutoModifier('prior-tfm') # for the generator
class Transformed(Prior):
	def __init__(self, A, **kwargs):

		super().__init__(A, **kwargs)

		prior_tfm = None
		if self.prior_dim is not None:

			tfm_info = A.pull('prior-tfm', None, silent=True, raw=True)
			if tfm_info is not None:
				A.push('prior-tfm.din', self.prior_dim, silent=True)
				A.push('prior-tfm.dout', self.prior_dim, silent=True)

			prior_tfm = A.pull('prior-tfm', None)
		self.prior_tfm = prior_tfm
		if self.prior_tfm is None:
			print('WARNING: not transforming prior')

	def sample_prior(self, N=1):
		q = super().sample_prior(N)
		return q if self.prior_tfm is None else self.prior_tfm(q)


# endregion


###############
# region Normal
###############

@fig.AutoModifier('normal')
class Normal(Function):
	def __init__(self, A, **kwargs):
		
		dout = None

		change_dout = A.pull('change-dout', True)
		if change_dout:
			dout_key = A.pull('_dout_key', None)
			if dout_key is None:
				dout = A.pull('_dout', None)
				if dout is not None:
					dout_key = '_dout'
				else:
					dout = A.pull('dout')
			else:
				dout = A.pull(dout_key)
			
			if dout is not None:

				if isinstance(dout, int):
					chn = dout * 2
					dout = chn

				else:
					chn = dout[0]
					chn = chn * 2
					dout = (chn, *dout[1:])
				
				if dout_key is None:
					A.push('dout', dout)
					# A.push('dout', dout, silent=True)
				else:
					A.push(dout_key, dout, silent=True)

		super().__init__(A, **kwargs)
		
		if dout is None:
			dout = self.dout
		self.full_dout = dout

		split = change_dout
		if not change_dout:
			split = A.pull('split-out', True)

		out = None
		if split:

			if isinstance(dout, int):
				assert dout % 2 == 0
				chn = dout // 2
				dout = chn

			else:
				chn = dout[0]
				assert chn % 2 == 0
				chn = chn // 2
				dout = (chn, *dout[1:])

		else:

			if isinstance(dout, int):
				chn = dout

				out = nn.Linear(chn, chn * 2)

			# A.push('out-layer._type', 'dense-layer', silent=True)
			# A.push('')

			else:
				chn = dout[0]

				out = nn.Conv2d(chn, chn * 2, kernel_size=1)

		# A.push('out-layer._type', 'conv-layer', silent=True)

		self.normal_layer = out
		self.width = chn
		self.dout = dout

	def forward(self, *args, **kwargs):

		x = super().forward(*args, **kwargs)

		if self.normal_layer is not None:
			x = self.normal_layer(x)

		mu, sigma = x.narrow(1, 0, self.width), x.narrow(1, self.width, self.width).exp()

		return NormalDistribution(mu, sigma)

# endregion




import sys, os

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models

from scipy import linalg

try:
    from torchvision.models.utils import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url


def calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
    """Numpy implementation of the Frechet Distance.
    The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
    and X_2 ~ N(mu_2, C_2) is
            d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).
    Stable version by Dougal J. Sutherland.
    Params:
    -- mu1   : Numpy array containing the activations of a layer of the
               inception net (like returned by the function 'get_predictions')
               for generated samples.
    -- mu2   : The sample mean over activations, precalculated on an
               representative data set.
    -- sigma1: The covariance matrix over activations for generated samples.
    -- sigma2: The covariance matrix over activations, precalculated on an
               representative data set.
    Returns:
    --   : The Frechet Distance.
    """

    mu1 = np.atleast_1d(mu1)
    mu2 = np.atleast_1d(mu2)

    sigma1 = np.atleast_2d(sigma1)
    sigma2 = np.atleast_2d(sigma2)

    assert mu1.shape == mu2.shape, \
        'Training and test mean vectors have different lengths'
    assert sigma1.shape == sigma2.shape, \
        'Training and test covariances have different dimensions'

    diff = mu1 - mu2

    # Product might be almost singular
    covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
    if not np.isfinite(covmean).all():
        msg = ('fid calculation produces singular product; '
               'adding %s to diagonal of cov estimates') % eps
        print(msg)
        offset = np.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

    # Numerical error might give slight imaginary component
    if np.iscomplexobj(covmean):
        if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
            m = np.max(np.abs(covmean.imag))
            raise ValueError('Imaginary component {}'.format(m))
        covmean = covmean.real

    tr_covmean = np.trace(covmean)

    return (diff.dot(diff) + np.trace(sigma1) +
            np.trace(sigma2) - 2 * tr_covmean)





# Inception weights ported to Pytorch from
# http://download.tensorflow.org/models/image/imagenet/inception-2015-12-05.tgz
FID_WEIGHTS_URL = 'https://github.com/mseitzer/pytorch-fid/releases/download/fid_weights/pt_inception-2015-12-05-6726825d.pth'

## Modified Inception Network for FID score
class InceptionV3(nn.Module):
  """Pretrained InceptionV3 network returning feature maps"""

  # Index of default block of inception to return,
  # corresponds to output of final average pooling
  DEFAULT_BLOCK_INDEX = 3

  # Maps feature dimensionality to their output blocks indices
  BLOCK_INDEX_BY_DIM = {
      64: 0,   # First max pooling features
      192: 1,  # Second max pooling featurs
      768: 2,  # Pre-aux classifier features
      2048: 3  # Final average pooling features
  }

  def __init__(self,
                 output_blocks=[DEFAULT_BLOCK_INDEX],
                 resize_input=True,
                 normalize_input=True,
                 requires_grad=False,
                 use_fid_inception=True):
    """Build pretrained InceptionV3
    Parameters
    ----------
    output_blocks : list of int
        Indices of blocks to return features of. Possible values are:
            - 0: corresponds to output of first max pooling
            - 1: corresponds to output of second max pooling
            - 2: corresponds to output which is fed to aux classifier
            - 3: corresponds to output of final average pooling
    resize_input : bool
        If true, bilinearly resizes input to width and height 299 before
        feeding input to model. As the network without fully connected
        layers is fully convolutional, it should be able to handle inputs
        of arbitrary size, so resizing might not be strictly needed
    normalize_input : bool
        If true, scales the input from range (0, 1) to the range the
        pretrained Inception network expects, namely (-1, 1)
    requires_grad : bool
        If true, parameters of the model require gradients. Possibly useful
        for finetuning the network
    use_fid_inception : bool
        If true, uses the pretrained Inception model used in Tensorflow's
        FID implementation. If false, uses the pretrained Inception model
        available in torchvision. The FID Inception model has different
        weights and a slightly different structure from torchvision's
        Inception model. If you want to compute FID scores, you are
        strongly advised to set this parameter to true to get comparable
        results.
    """
    super(InceptionV3, self).__init__()

    self.resize_input = resize_input
    self.normalize_input = normalize_input
    self.output_blocks = sorted(output_blocks)
    self.last_needed_block = max(output_blocks)

    assert self.last_needed_block <= 3, \
        'Last possible output block index is 3'

    self.blocks = nn.ModuleList()

    if use_fid_inception:
        inception = fid_inception_v3()
    else:
        inception = models.inception_v3(pretrained=True)

    # Block 0: input to maxpool1
    block0 = [
        inception.Conv2d_1a_3x3,
        inception.Conv2d_2a_3x3,
        inception.Conv2d_2b_3x3,
        nn.MaxPool2d(kernel_size=3, stride=2)
    ]
    self.blocks.append(nn.Sequential(*block0))

    # Block 1: maxpool1 to maxpool2
    if self.last_needed_block >= 1:
        block1 = [
            inception.Conv2d_3b_1x1,
            inception.Conv2d_4a_3x3,
            nn.MaxPool2d(kernel_size=3, stride=2)
        ]
        self.blocks.append(nn.Sequential(*block1))

    # Block 2: maxpool2 to aux classifier
    if self.last_needed_block >= 2:
        block2 = [
            inception.Mixed_5b,
            inception.Mixed_5c,
            inception.Mixed_5d,
            inception.Mixed_6a,
            inception.Mixed_6b,
            inception.Mixed_6c,
            inception.Mixed_6d,
            inception.Mixed_6e,
        ]
        self.blocks.append(nn.Sequential(*block2))

    # Block 3: aux classifier to final avgpool
    if self.last_needed_block >= 3:
        block3 = [
            inception.Mixed_7a,
            inception.Mixed_7b,
            inception.Mixed_7c,
            nn.AdaptiveAvgPool2d(output_size=(1, 1))
        ]
        self.blocks.append(nn.Sequential(*block3))

    for param in self.parameters():
        param.requires_grad = requires_grad

  def forward(self, inp):
    """Get Inception feature maps
    Parameters
    ----------
    inp : torch.autograd.Variable
        Input tensor of shape Bx3xHxW. Values are expected to be in
        range (0, 1)
    Returns
    -------
    List of torch.autograd.Variable, corresponding to the selected output
    block, sorted ascending by index
    """
    outp = []
    x = inp

    if self.resize_input:
        x = F.interpolate(x,
                          size=(299, 299),
                          mode='bilinear',
                          align_corners=False)

    if self.normalize_input:
        x = 2 * x - 1  # Scale from range (0, 1) to range (-1, 1)

    for idx, block in enumerate(self.blocks):
        x = block(x)
        if idx in self.output_blocks:
            outp.append(x)

        if idx == self.last_needed_block:
            break

    return outp

class FIDInceptionA(models.inception.InceptionA):
  """InceptionA block patched for FID computation"""
  def __init__(self, in_channels, pool_features):
    super(FIDInceptionA, self).__init__(in_channels, pool_features)

  def forward(self, x):
    branch1x1 = self.branch1x1(x)

    branch5x5 = self.branch5x5_1(x)
    branch5x5 = self.branch5x5_2(branch5x5)

    branch3x3dbl = self.branch3x3dbl_1(x)
    branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl = self.branch3x3dbl_3(branch3x3dbl)

    # Patch: Tensorflow's average pool does not use the padded zero's in
    # its average calculation
    branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                               count_include_pad=False)
    branch_pool = self.branch_pool(branch_pool)

    outputs = [branch1x1, branch5x5, branch3x3dbl, branch_pool]
    return torch.cat(outputs, 1)


class FIDInceptionC(models.inception.InceptionC):
  """InceptionC block patched for FID computation"""
  def __init__(self, in_channels, channels_7x7):
    super(FIDInceptionC, self).__init__(in_channels, channels_7x7)

  def forward(self, x):
    branch1x1 = self.branch1x1(x)

    branch7x7 = self.branch7x7_1(x)
    branch7x7 = self.branch7x7_2(branch7x7)
    branch7x7 = self.branch7x7_3(branch7x7)

    branch7x7dbl = self.branch7x7dbl_1(x)
    branch7x7dbl = self.branch7x7dbl_2(branch7x7dbl)
    branch7x7dbl = self.branch7x7dbl_3(branch7x7dbl)
    branch7x7dbl = self.branch7x7dbl_4(branch7x7dbl)
    branch7x7dbl = self.branch7x7dbl_5(branch7x7dbl)

    # Patch: Tensorflow's average pool does not use the padded zero's in
    # its average calculation
    branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                               count_include_pad=False)
    branch_pool = self.branch_pool(branch_pool)

    outputs = [branch1x1, branch7x7, branch7x7dbl, branch_pool]
    return torch.cat(outputs, 1)


class FIDInceptionE_1(models.inception.InceptionE):
  """First InceptionE block patched for FID computation"""
  def __init__(self, in_channels):
    super(FIDInceptionE_1, self).__init__(in_channels)

  def forward(self, x):
    branch1x1 = self.branch1x1(x)

    branch3x3 = self.branch3x3_1(x)
    branch3x3 = [
        self.branch3x3_2a(branch3x3),
        self.branch3x3_2b(branch3x3),
    ]
    branch3x3 = torch.cat(branch3x3, 1)

    branch3x3dbl = self.branch3x3dbl_1(x)
    branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl = [
        self.branch3x3dbl_3a(branch3x3dbl),
        self.branch3x3dbl_3b(branch3x3dbl),
    ]
    branch3x3dbl = torch.cat(branch3x3dbl, 1)

    # Patch: Tensorflow's average pool does not use the padded zero's in
    # its average calculation
    branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                               count_include_pad=False)
    branch_pool = self.branch_pool(branch_pool)

    outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
    return torch.cat(outputs, 1)


class FIDInceptionE_2(models.inception.InceptionE):
  """Second InceptionE block patched for FID computation"""
  def __init__(self, in_channels):
    super(FIDInceptionE_2, self).__init__(in_channels)

  def forward(self, x):
    branch1x1 = self.branch1x1(x)

    branch3x3 = self.branch3x3_1(x)
    branch3x3 = [
        self.branch3x3_2a(branch3x3),
        self.branch3x3_2b(branch3x3),
    ]
    branch3x3 = torch.cat(branch3x3, 1)

    branch3x3dbl = self.branch3x3dbl_1(x)
    branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl = [
        self.branch3x3dbl_3a(branch3x3dbl),
        self.branch3x3dbl_3b(branch3x3dbl),
    ]
    branch3x3dbl = torch.cat(branch3x3dbl, 1)

    # Patch: The FID Inception model uses max pooling instead of average
    # pooling. This is likely an error in this specific Inception
    # implementation, as other Inception models use average pooling here
    # (which matches the description in the paper).
    branch_pool = F.max_pool2d(x, kernel_size=3, stride=1, padding=1)
    branch_pool = self.branch_pool(branch_pool)

    outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
    return torch.cat(outputs, 1)

def fid_inception_v3():
  """Build pretrained Inception model for FID computation
  The Inception model for FID computation uses a different set of weights
  and has a slightly different structure than torchvision's Inception.
  This method first constructs torchvision's Inception and then patches the
  necessary parts that are different in the FID Inception model.
  """
  inception = models.inception_v3(num_classes=1008,
                                  aux_logits=False,
                                  pretrained=False,
                                  init_weights=False)
  inception.Mixed_5b = FIDInceptionA(192, pool_features=32)
  inception.Mixed_5c = FIDInceptionA(256, pool_features=64)
  inception.Mixed_5d = FIDInceptionA(288, pool_features=64)
  inception.Mixed_6b = FIDInceptionC(768, channels_7x7=128)
  inception.Mixed_6c = FIDInceptionC(768, channels_7x7=160)
  inception.Mixed_6d = FIDInceptionC(768, channels_7x7=160)
  inception.Mixed_6e = FIDInceptionC(768, channels_7x7=192)
  inception.Mixed_7b = FIDInceptionE_1(1280)
  inception.Mixed_7c = FIDInceptionE_2(2048)

  state_dict = load_state_dict_from_url(FID_WEIGHTS_URL, progress=True)
  inception.load_state_dict(state_dict)
  return inception







import math
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils import spectral_norm

from omnibelt import unspecified_argument
import omnifig as fig

# from .. import framework as fm
from ..op import framework as fm
from .. import util


#################
# region Functions
#################

def make_MLP(din, dout, hidden=None,
			 initializer=None,
			 nonlin='elu', output_nonlin=None,
			 logify_in=False, unlogify_out=False,
			 bias=True, output_bias=None):
	'''
	:param din: int
	:param dout: int
	:param hidden: ordered list of int - each element corresponds to a FC layer with that width (empty means network is not deep)
	:param nonlin: str - choose from options found in get_nonlinearity(), applied after each intermediate layer
	:param output_nonlin: str - nonlinearity to be applied after the last (output) layer
	:param logify_in: convert input to logified space
	:param unlogify_out: convert output back to linear space
	:return: an nn.Sequential instance with the corresponding layers
	'''

	if hidden is None:
		hidden = []

	if output_bias is None:
		output_bias = bias

	flatten = False
	reshape = None

	din = din
	dout = dout

	if isinstance(din, (tuple, list)):
		flatten = True
		din = int(np.product(din))
	if isinstance(dout, (tuple, list)):
		reshape = dout
		dout = int(np.product(dout))

	nonlins = [nonlin] * len(hidden) + [output_nonlin]
	biases = [bias] * len(hidden) + [output_bias]
	hidden = din, *hidden, dout

	layers = []
	if flatten:
		layers.append(nn.Flatten())

	if logify_in:
		layers.append(util.Logifier())
	for in_dim, out_dim, nonlin, bias in zip(hidden, hidden[1:], nonlins, biases):
		layer = nn.Linear(in_dim, out_dim, bias=bias)
		if initializer is not None:
			layer = initializer(layer, nonlin)
		layers.append(layer)
		if nonlin is not None:
			layers.append(util.get_nonlinearity(nonlin))

	if unlogify_out:
		layers.append(util.Unlogifier())
	if reshape is not None:
		layers.append(Reshaper(reshape))


	net = nn.Sequential(*layers)

	net.din, net.dout = din, dout
	return net

def batched_grouped_linear(points, weights, masks, biases=None): # equivalent to Ntfm
	'''
	Linear layer where the weights (and biases) are batched and grouped as well,
	so the output is a weighted sum of the outputs from each group

	Use case: points=point_cloud, weights=SE3_rotations, masks=segmentation_masks, biases=SE_translation

	points: (B, C, M)
	weights: (B, N, D, C)
	biases: (B, N, D)
	masks: (B, N, M) - sum across N == 1

	out: (B, D, M)
	'''

	features = weights @ points.unsqueeze(1)  # (B, N, D, M)

	#print(points.shape, weights.shape, masks.shape, biases.shape, features.shape)

	if biases is not None:
		features += biases#.unsqueeze(-1)

	out = features.mul(masks.unsqueeze(2)).sum(1)

	return out

fig.AutoComponent('flatten')(nn.Flatten)

@fig.AutoComponent('reshaper')
class Reshaper(nn.Module): # by default flattens

	def __init__(self, dout=(-1,)):
		super().__init__()

		self.dout = dout

	def extra_repr(self):
		return f'out={self.dout}'

	def forward(self, x):
		B = x.size(0)
		return x.view(B, *self.dout)

@fig.Modification('spec-norm')
def spec_norm_layer(layer, config):
	
	kwargs = dict(
		name = config.pull('weight_name', 'weight'),
		n_power_iterations = config.pull('n_power_iterations', 1),
		eps = config.pull('eps', 1e-12),
		dim = config.pull('dim', None),
	)
	
	return spectral_norm(layer, **kwargs)

# endregion
#################
# region Layers
#################

@fig.AutoComponent('recurrence')
class Recurrence(fm.FunctionBase):
	def __init__(self, din, dout, hidden_dim=None, num_layers=1, output_nonlin=None,
				 rec_type='gru', dropout=0., batch_first=False, bidirectional=False, auto_reset=True):
		super().__init__(din, dout)

		assert rec_type in {'gru', 'lstm', 'rnn'}

		self.rec_dim = hidden_dim
		if hidden_dim is None:
			self.rec_dim = dout

		rec_args = {'input_size': din, 'hidden_size': self.rec_dim, 'num_layers': num_layers,
					'dropout': dropout, 'batch_first': batch_first, 'bidirectional': bidirectional}
		if rec_type == 'rnn':
			self.rec = nn.RNN(**rec_args)
		elif rec_type == 'rnn-relu':
			self.rec = nn.RNN(nonlinearity='relu', **rec_args)
		elif rec_type == 'gru':
			self.rec = nn.GRU(**rec_args)
		elif rec_type == 'lstm':
			self.rec = nn.LSTM(**rec_args)

		self.out_layer = None
		self.out_nonlin = None
		if hidden_dim is not None:
			self.out_layer = make_MLP(self.rec_dim, dout, output_nonlin=output_nonlin)

		self.auto_reset = auto_reset
		self.reset()

	def reset(self, hidden=None):
		self.hidden = hidden

	def forward(self, seq, hidden=None, ret_hidden=False):  # input dims: seq x batch x input

		if self.auto_reset:
			self.reset(hidden)
		hidden = self.hidden if hidden is None else hidden

		out, self.hidden = self.rec(seq, hidden)  # if self.hidden is None else self.rec(seq, self.hidden)

		if self.out_layer is not None:
			out = self.out_layer(out)

		if ret_hidden:
			return out, self.hidden

		return out

@fig.AutoComponent('fourier-layer')
class FourierLayer(fm.FunctionBase): # TODO: generalize period to multiple din
	def __init__(self, din=1, dout=1, order=100, periods=None):
		super().__init__(din, dout)

		self.sin_coeff = nn.Linear(self.din * order, self.dout, bias=False)
		self.cos_coeff = nn.Linear(self.din * order, self.dout, bias=True)
		self.order = order

		if periods is not None:
			w = 2* np.pi / periods
			self.register_buffer('w', w.view(1, -1))
		else:
			w = nn.Parameter(torch.tensor([2 * np.pi] * din).view(1, -1), requires_grad=True)
			self.register_parameter('w', w)

		freqs = torch.arange(1, order + 1).view(1, 1, -1).float()  # 1 x 1 x O
		self.register_buffer('freqs', freqs)

	def forward(self, x):

		x = x.view(-1, self.din).mul(self.w).unsqueeze(-1)  # B x D x 1
		x = x.mul(self.freqs).view(-1, self.din * self.order)

		return self.sin_coeff(torch.sin(x)) + self.cos_coeff(torch.cos(x))

	def get_period(self):
		return 2 * np.pi / self.w


class Invertible(fm.FunctionBase):
	def __init__(self, features, bias=True):
		super().__init__(features, features)
		
		self.features = features
		
		self.wU = nn.Parameter(torch.Tensor(features, features))
		self.wV = nn.Parameter(torch.Tensor(features, features))
		self.log_S = nn.Parameter(torch.Tensor(features))
		
		if bias:
			self.bias = nn.Parameter(torch.Tensor(features))
		else:
			self.register_parameter('bias', None)
		self.reset_parameters()
	
	def reset_parameters(self):
		nn.init.kaiming_uniform_(self.wU, a=math.sqrt(5))
		nn.init.kaiming_uniform_(self.wV, a=math.sqrt(5))
		
		fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.wV)
		bound = 1 / math.sqrt(fan_in)
		nn.init.uniform_(self.log_S, -bound, bound)
		
		if self.bias is not None:
			nn.init.uniform_(self.bias, -bound, bound)
		
	def _svd(self, inverse=False):
		U = util.orthogonalize(self.wU)
		V = util.orthogonalize(self.wV)
		log_S = self.log_S
		
		if inverse:
			V, U = U.t(), V.t()
			log_S = -log_S
			
		return U, log_S.exp(), V
		
	def _mat(self, inverse=False):
		U, S, V = self._svd(inverse=inverse)
		return (U * S.unsqueeze(0)) @ V
		
		
	def forward(self, input):
		return F.linear(input, self._mat(), self.bias)
		
	def inverse(self, input):
		if self.bias is not None:
			input = input - self.bias.unsqueeze(0)
		return F.linear(input, self._mat(True))
	
	def extra_repr(self):
		return f'features={self.features}, bias={self.bias is not None}'

@fig.AutoComponent('inv-layer')
class InvertibleLayer(Invertible):
	def __init__(self, A):
		dim = A.pull('dim', '<>features', None)
		if dim is None:
			dim = A.pull('din', None)
		if dim is None:
			dim = A.pull('dout', None)
		if dim is None:
			raise Exception('unknown size')
		
		bias = A.pull('bias', True)
		
		super().__init__(dim, bias=bias)


class PowerLinear(fm.FunctionBase):  # includes powers of input as features
	def __init__(self):
		raise NotImplementedError


class DenseLayerBase(fm.FunctionBase):
	def __init__(self, din, dout, bias=True, nonlin='elu', **kwargs):
		super().__init__(din=din, dout=dout, **kwargs)

		seq = []

		if isinstance(din, (tuple, list)):
			seq.append(nn.Flatten())
			din = int(np.product(din))

		_dout = int(np.product(dout)) if isinstance(dout, (tuple, list)) else dout

		seq.append(nn.Linear(din, _dout, bias=bias))

		nonlin = util.get_nonlinearity(nonlin)
		if nonlin is not None:
			seq.append(nonlin)

		if isinstance(dout, (tuple, list)):
			seq.append(Reshaper(dout))

		self.layer = nn.Sequential(*seq) if len(seq) > 1 else seq[0]

	def forward(self, x):
		return self.layer(x)


@fig.Component('dense-layer')
class DenseLayer(fm.Function, DenseLayerBase):
	def __init__(self, A, din=None, dout=None, bias=None, nonlin=None, **kwargs):
		
		if din is None:
			din = A.pull('din')
			if din is None:
				din = A.pull('width', '<>channels')
		if dout is None:
			dout = A.pull('dout')
			if dout is None:
				dout = A.pull('width', '<>channels')
		if bias is None:
			bias = A.pull('bias', True)
		if nonlin is None:
			nonlin = A.pull('nonlin', 'elu')
		
		super().__init__(A, din=din, dout=dout, bias=bias, nonlin=nonlin, **kwargs)


class ConvLayerBase(fm.FunctionBase):
	def __init__(self, cin, cout, kernel_size,
	             padding=None, stride=1, dilation=1, output_padding=0,
	             unpool=None, pool=None, norm=None, nonlin=None,
	             residual=False, force_res=False, transpose=False, print_dims=True,
	             conv_kwargs={}, **kwargs):
		
		super().__init__(**kwargs)
		
		self.unpool = unpool
		
		if transpose:
			self.conv = nn.ConvTranspose2d(cin, cout, kernel_size=kernel_size, padding=padding,
			                               stride=stride, dilation=dilation, output_padding=output_padding,
			                               **conv_kwargs)
		else:
			self.conv = nn.Conv2d(cin, cout, kernel_size=kernel_size, padding=padding,
			                      stride=stride, dilation=dilation, **conv_kwargs)
		
		self.pool = pool
		self.norm = norm
		self.nonlin = util.get_nonlinearity(nonlin)
		
		self.res = (force_res or (residual and cin == cout)) \
		           and (self.conv.stride[0] == self.conv.stride[1] == 1)
		if force_res and not self.res:
			print(f'WARNING: cant use residual due to stride {stride}')
		if self.res and (cin != cout):
			print('WARNING: residual connections will be partial, because channels dont match')
		
		self.print_dims = print_dims
	
	def extra_repr(self):
		msg = ''
		if self.print_dims:
			msg += f'{self.din} -> {self.dout}\n'
		msg += f'residual={self.res}'
		return msg
	
	def forward(self, x):
		if self.unpool is not None:
			x = self.unpool(x)
		c = self.conv(x)
		if self.res:
			din, dout = x.size(1), c.size(1)
			if din > dout:
				x = c + x.narrow(1, 0, dout)
			else:
				if din < dout:
					B, _, H, W = x.size()
					x = torch.cat([x, torch.zeros(B, dout - din, H, W, device=x.device)], dim=1)
				
				x = c + x
		else:
			x = c
		
		if self.pool is not None:
			x = self.pool(x)
		if self.norm is not None:
			x = self.norm(x)
		if self.nonlin is not None:
			x = self.nonlin(x)
		return x




@fig.Component('conv-layer')
class ConvLayer(fm.Function, ConvLayerBase):

	def __init__(self, A, din=None, dout=None, channels=None,
	             down=None, up=None, size=None, transpose=None,
	             **kwargs):

		if din is None:
			din = A.pull('in-shape', '<>din', None)
		if channels is None:
			channels = A.pull('channels', None)
		if dout is None:
			dout = A.pull('out-shape', '<>dout', None)
			
		if transpose is None:
			transpose = A.pull('transpose', None)

		assert din is not None or dout is not None, 'no input info'

		if din is not None and dout is not None:

			Cin, Hin, Win = din
			Cout, Hout, Wout = dout

			if channels is None:
				channels = Cout

			if Hin != Hout or Win != Wout:
				H, W = Hout/Hin, Wout/Win

				val = None
				is_down = not ((H >= 1 and W >= 1) or (H < 1 and W > 1) or (H > 1 and W < 1))
				if is_down:
					H, W = 1/H, 1/W
				if int(H) == H and int(W) == W:
					val = int(H), int(W)
				else:
					size = Hout, Wout
				down = val if is_down else None
				up = None if is_down else val
		else:
			if down is None:
				down = A.pull('down', down)
			if isinstance(down, int):
				down = down, down
			assert down is None or (down[0] >= 1 and down[1] >= 1), f'{down}'
			if up is None and transpose is not False:
				up = A.pull('up', up) #if down is None else None
			if isinstance(up, int):
				up = up, up
			if down is None and up is not None and up[0] == up[1] == 1 and transpose is None:
				down = up
				up = None
			assert up is None or (up[0] >= 1 and up[1] >= 1), f'{up}'
			if down is None and up is None:
				if size is None:
					size = A.pull('size', None)
				if size is None:
					down = (1,1)
					
			if down is not None and up is not None:
				down = down if not transpose else None
				up = up if transpose else None

		pool = None if down is None or (down[0] == down[1] == 1) \
			else util.get_pooling(A.pull('pool', None), down)
		# A.begin()
		A.push('channels', channels, silent=True)
		unpool = util.get_upsample(A.pull('unpool', None), size=size) if size is not None \
			else (util.get_upsample(A.pull('unpool', None), up=up, channels=channels)
			      if up is not None and (up[0] > 1 or up[1] > 1) else None)
		# A.abort()
		if transpose: # make sure transpose is valid
			transpose = down is None or up is not None
		is_deconv = size is None and unpool is None and (transpose or (up is not None and down is None))
		# is_deconv = down is None and size is None and unpool is None

		kernel_size = A.pull('kernel_size', '<>kernel', (3+up[0]-1, 3+up[1]-1) if is_deconv else (3,3)) # TODO: check condition for larger default kernel for deconvs
		if kernel_size is None:
			kernel_size = (4,4) if is_deconv else (3,3)
		if isinstance(kernel_size, int):
			kernel_size = kernel_size, kernel_size
		padding = A.pull('padding', (kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2)
		if padding is None:
			padding = (kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2
		if isinstance(padding, int):
			padding = padding, padding
		dilation = A.pull('dilation', (1,1))
		if dilation is None:
			dilation = (1,1)
		if isinstance(dilation, int):
			dilation = dilation, dilation
		output_padding = A.pull('output_padding', (0,0)) if is_deconv else None
		if isinstance(output_padding, int):
			output_padding = output_padding, output_padding

		if is_deconv:
			stride = up
		elif down is not None and pool is None:
			stride = down
		else:
			stride = (1,1)

		_stride = stride
		stride = A.pull('stride', stride)
		if stride is None:
			stride = _stride
		if isinstance(stride, int):
			stride = stride, stride

		if din is None:
			Cout, Hout, Wout = dout
			if channels is None:
				channels = Cout

			Cin = channels

			H, W = Hout, Wout

			if down is not None and pool is not None:
				H, W = H*down[0], W*down[1]

			if is_deconv:
				if output_padding is not None:
					H, W = H - output_padding[0], W - output_padding[1]
				H, W = util.conv_size_change(H, W, kernel_size, padding, stride, dilation)
			else:
				H, W = util.deconv_size_change(H, W, kernel_size, padding, stride, dilation)

			if up is not None and unpool is not None:
				H, W = H//up[0], W//up[1]

			if size is not None:
				assert size == (H, W), f'{size} vs {(H, W)}'
				H, W = None, None
				raise NotImplementedError

			Hin, Win = H, W

		if dout is None:
			Cin, Hin, Win = din
			if channels is None:
				channels = Cin

			H, W = Hin, Win

			if size is not None:
				H, W = size

			if up is not None and unpool is not None:
				H, W = H*up[0], W*up[1]

			Cout = channels

			if is_deconv:
				H, W = util.deconv_size_change(H, W, kernel_size, padding, stride, dilation,
													 output_padding)
			else:
				H, W = util.conv_size_change(H, W, kernel_size, padding, stride, dilation)

			if down is not None and pool is not None:
				H, W = H//down[0], W//down[1]

			Hout, Wout = H, W

		# A.begin()
		A.push('channels', Cout, silent=True)
		norm = util.get_normalization(A.pull('norm', None), Cout)
		# A.abort()

		nonlin = util.get_nonlinearity(A.pull('nonlin', 'elu'))

		conv_kwargs = A.pull('conv_kwargs', {})

		residual = A.pull('residual', False)
		force_res = A.pull('force_res', False)

		print_dims = A.pull('print_dims', False)
		din, dout = (Cin, Hin, Win), (Cout, Hout, Wout)

		super().__init__(A, din=din, dout=dout,
		                 cin=Cin, cout=Cout, kernel_size=kernel_size,
	             padding=padding, stride=stride, dilation=dilation, output_padding=output_padding,
	             unpool=unpool, pool=pool, norm=norm, nonlin=nonlin,
	             residual=residual, force_res=force_res, transpose=is_deconv, print_dims=print_dims,
	             conv_kwargs=conv_kwargs, **kwargs)


class ExpandedConvLayer(ConvLayer):
	def __init__(self, A, din=None, dout=None, channels=None, extra_channels=0, **kwargs):
		if din is None:
			din = A.pull('in-shape', '<>din', None)
		if channels is None:
			channels = A.pull('channels', None)
		if dout is None:
			dout = A.pull('out-shape', '<>dout', None)
		
		assert din is not None or dout is not None, 'no input info'
		
		if din is not None and dout is not None:
			Cin, Hin, Win = din
			Cout, Hout, Wout = dout
			if channels is None:
				channels = Cout
		
		if din is not None:
			Cin, Hin, Win = din
			Cin += extra_channels
			din = Cin, Hin, Win
		else:
			if channels is None:
				channels = dout[0]
			channels += extra_channels
		
		super().__init__(A, din=din, dout=dout, channels=channels, **kwargs)
		
	def preprocess(self, x):
		return x
		
	def forward(self, x):
		x = self.preprocess(x)
		return super().forward(x)
	
	
@fig.Component('coord-conv-layer')
class CoordConvLayer(ExpandedConvLayer):
	def __init__(self, A, din=None, dout=None, channels=None, extra_channels=0, **kwargs):
		super().__init__(A, din=din, dout=dout, channels=channels,
		                 extra_channels=extra_channels+2, **kwargs)
		
		C, H, W = self.din
		Hax = torch.linspace(0, 1, H)
		Wax = torch.linspace(0, 1, W)
		coords = torch.stack(torch.meshgrid(Hax, Wax), 0)
		self.register_buffer('coords', coords)
	
	def preprocess(self, x):
		x = super().preprocess(x)
		coords = self.coords.unsqueeze(0).expand(x.size(0), *self.coords.shape)
		return torch.cat([x, coords], 1)


@fig.AutoComponent('layer-norm')
class LayerNorm(fm.FunctionBase):
	def __init__(self, din, eps=1e-5, elementwise_affine=True):
		super().__init__(din, din)
		self.norm = nn.LayerNorm(din, eps=eps,
				elementwise_affine=elementwise_affine)

	def forward(self, x):
		return self.norm(x)


@fig.AutoComponent('interpolate')
class Interpolate(fm.FunctionBase):
	def __init__(self, din=None, size=None, scale_factor=None, mode='nearest',
				 align_corners=None, recompute_scale_factor=None):
		assert size is not None or (din is not None and scale_factor is not None)

		Cin = None
		if din is not None:
			Cin, Hin, Win = din

			if size is None:
				size = int(scale_factor*Hin), int(scale_factor*Win)

		super().__init__(din, (Cin, *size))

		self.size = size
		self.scale_factor = scale_factor
		self.mode = mode
		self.align_corners = align_corners
		self.recompute_scale_factor = recompute_scale_factor

	def forward(self, x):
		return F.interpolate(x, size=self.size, scale_factor=self.scale_factor,
							 mode=self.mode, align_corners=self.align_corners,
							 recompute_scale_factor=self.recompute_scale_factor)


@fig.AutoComponent('conv-lstm')
class ConvLSTM(fm.FunctionBase):
	def __init__(self, input_dim, hidden_dim, kernel_size,
				 batch_first=False, auto_reset=True,
				 peephole=True):
		super().__init__(input_dim, hidden_dim)

		self.kernel_size = kernel_size
		self.padding = int((kernel_size - 1) / 2)

		self.Wxi = nn.Conv2d(self.din, self.dout, self.kernel_size, 1, self.padding, bias=True)
		self.Whi = nn.Conv2d(self.dout, self.dout, self.kernel_size, 1, self.padding, bias=False)
		self.Wxf = nn.Conv2d(self.din, self.dout, self.kernel_size, 1, self.padding, bias=True)
		self.Whf = nn.Conv2d(self.dout, self.dout, self.kernel_size, 1, self.padding, bias=False)
		self.Wxc = nn.Conv2d(self.din, self.dout, self.kernel_size, 1, self.padding, bias=True)
		self.Whc = nn.Conv2d(self.dout, self.dout, self.kernel_size, 1, self.padding, bias=False)
		self.Wxo = nn.Conv2d(self.din, self.dout, self.kernel_size, 1, self.padding, bias=True)
		self.Who = nn.Conv2d(self.dout, self.dout, self.kernel_size, 1, self.padding, bias=False)

		if peephole: # pixelwise
			param = util.create_param(1, self.dout, 1, 1)
			param.data.zero_()
			self.register_parameter('Wci', param)
			param = util.create_param(1, self.dout, 1, 1)
			param.data.zero_()
			self.register_parameter('Wcf', param)
			param = util.create_param(1, self.dout, 1, 1)
			param.data.zero_()
			self.register_parameter('Wco', param)
		else:
			self.Wci = 0
			self.Wcf = 0
			self.Wco = 0

		self.register_buffer('hidden', torch.zeros(1, self.dout, 1, 1))
		self.register_buffer('cell', torch.zeros(1, self.dout, 1, 1))

		self.auto_reset = auto_reset
		self.batch_first = int(batch_first)

	def reset(self):
		self.hidden = torch.zeros(*((1,) + self.hidden.shape[1:]), dtype=self.hidden.dtype, device=self.hidden.device)
		self.cell = torch.zeros(*((1,) + self.cell.shape[1:]), dtype=self.cell.dtype, device=self.cell.device)

	def seq(self, xs, h=None, c=None, ret_cell=False):

		if self.batch_first:
			xs = xs.transpose(0,1)

		outs = []
		hiddens = []

		for x in xs:
			h, c = self(x,h,c, ret_cell=True)
			outs.append(h)
			hiddens.append(c)

		if self.auto_reset:
			self.reset()

		outs = torch.stack(outs, self.batch_first)
		if ret_cell:
			cells = torch.stack(outs, self.batch_first)
			return outs, cells
		return outs

	def forward(self, x, h=None, c=None, ret_cell=False):

		B, C, H, W = x.shape

		if h is None:
			if (H,W) != self.hidden.shape[-2:]:
				self.hidden = torch.zeros(1, self.dout, H, W, dtype=x.dtype, device=x.device)
			h = self.hidden

		if c is None:
			if (H,W) != self.cell.shape[-2:]:
				self.cell = torch.zeros(1, self.dout, H, W, dtype=x.dtype, device=x.device)
			c = self.cell

		ci = torch.sigmoid(self.Wxi(x) + self.Whi(h) + c * self.Wci)
		cf = torch.sigmoid(self.Wxf(x) + self.Whf(h) + c * self.Wcf)
		cc = cf * c + ci * torch.tanh(self.Wxc(x) + self.Whc(h))
		co = torch.sigmoid(self.Wxo(x) + self.Who(h) + cc * self.Wco)
		ch = co * torch.tanh(cc)

		self.hidden = ch
		self.cell = cc

		if ret_cell:
			return ch, cc
		return ch

# endregion
#################





name = 'omnilearn'
long_name = 'omni-learn'

version = '0.5.4'
url = 'https://github.com/felixludos/omni-learn'

description = 'Powerful machine learning utilities for python'

author = 'Felix Leeb'
author_email = 'felixludos.info@gmail.com'

license = 'MIT'

readme = 'README.md'

installable_packages = ['omnilearn']

import os
try:
	with open(os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))), 'requirements.txt'), 'r') as f:
		install_requires = f.readlines()
except:
	install_requires = ['humpack', 'omnibelt', 'omnifig', 'pyyaml', 'c3linearize', 'numpy', 'matplotlib',
	                    'torch', 'tensorflow',
	                    'gym', 'tabulate', 'ipdb', 'h5py', 'pyyaml', 'tqdm', 'pandas']
del os




import torch

import omnifig as fig

from .nets import MultiLayer

from ..op.framework import Function
from .. import util

from .features import Prior, Gaussian, Uniform

class StyleFusionLayer(Function):
	def __init__(self, A, style_dim=None, **kwargs):

		if style_dim is None:
			style_dim = A.pull('style-dim', None)
			
		super().__init__(A, **kwargs)

		self.style_dim = style_dim
		self.register_buffer('_style', None)

	def get_style_dim(self):
		return self.style_dim

	def clear_style(self):
		self.cache_style(None)

	def cache_style(self, style):
		if self._style is None:
			self._style = style

	def process_style(self, style):

		if style is None:
			assert self._style is not None, 'no style provided or cached'
			style = self._style
		elif self._style is not None:
			raise Exception('Provided a style, when one was already cached')

		self._style = None

		return style
	
	def process_content(self, content):
		return content

	def infuse(self, content, style, **kwargs):
		raise NotImplementedError

	def forward(self, content, style=None, **kwargs):

		style = self.process_style(style)
		content = self.process_content(content)

		return self.infuse(content, style, **kwargs)


class PriorStyleFusionLayer(Prior, StyleFusionLayer):

	def __init__(self, A, style_dim=None, **kwargs):
		super().__init__(A, style_dim=style_dim, prior_dim=style_dim, **kwargs)
		if style_dim is None:
			self.prior_dim = self.style_dim

	def forward(self, content, style=None, **kwargs):
		if style is None and self._style is None:
			self.cache_style(self.sample_prior(content.size(0)))
		return super().forward(content, style=style, **kwargs)


@fig.AutoModifier('gaussian-style')
class Gaussian(Gaussian, PriorStyleFusionLayer):
	pass

@fig.AutoModifier('uniform-style')
class Uniform(Uniform, PriorStyleFusionLayer):
	pass


class StyleExtractorLayer(Function):
	def __init__(self, A, style_dim=None, ret_content=None, ret_style=None, **kwargs):
		if style_dim is None:
			style_dim = A.pull('style-dim', None)

		if ret_content is None:
			ret_content = A.pull('ret_content', True)
		if ret_style is None:
			ret_style = A.pull('ret_style', False)

		super().__init__(A, **kwargs)

		self.style_dim = style_dim
		self.register_buffer('_style', None)
		self.register_buffer('_content', None)
		
		self._ret_style = ret_style
		self._ret_content = ret_content
		
	def get_style_dim(self):
		return self.style_dim
	
	def process_style(self, style):
		return style
	
	def process_content(self, content):
		return content
	
	def clear_cache(self, content=True, style=True):
		if content:
			self.cache_content()
		if style:
			self.cache_style()
	
	def cache_style(self, style=None):
		self._style = style
		
	def cache_content(self, content=None):
		self._content = content
		
	def collect_style(self):
		style = self._style
		self._style = None
		return style
	
	def extract(self, inp, **unused):
		raise NotImplementedError
	
	def forward(self, inp, ret_style=None, ret_content=None, **kwargs):
		if ret_style is None:
			ret_style = self._ret_style
		if ret_content is None:
			ret_content = self._ret_content
		
		content, style = self.extract(inp, **kwargs)
		
		content = self.process_content(content)
		style = self.process_style(style)

		if ret_content and ret_style:
			return content, style
		
		if ret_style:
			self.cache_content(content)
			return style
		
		self.cache_style(style)
		return content
		

class StyleMultiLayer(MultiLayer):
	
	def clear_style(self):
		super().clear_style()
		for layer in self.style_layers:
			layer.clear_style()


@fig.Component('style-fusion')
class StyleFusion(StyleFusionLayer, StyleMultiLayer):
	def __init__(self, A, **kwargs):

		split_style = A.pull('split-style', False)

		super().__init__(A, **kwargs)

		self.style_layers = [layer for layer in self.layers if isinstance(layer, StyleFusionLayer)]

		style_dims = None
		if split_style:
			style_dims = [layer.get_style_dim() for layer in self.style_layers]
			style_dim = sum(style_dims)
		else:
			style_dim = self.style_layers[0].get_style_dim()
			for layer in self.style_layers:
				assert layer.get_style_dim() == style_dim

		self.style_dims = style_dims
		self.style_dim = style_dim


	def share_style(self, style):
		styles = [style] * len(self.style_layers) if self.style_dims is None else style.split(self.style_dims, dim=1)
		for layer, style in zip(self.style_layers, styles):
			layer.cache_style(style)
		return styles
	
	
	def infuse(self, content, style, **unused):
		
		self.share_style(style)

		# run forward pass through all layers starting with "root"
		return super(StyleFusionLayer, self).forward(content)



@fig.Component('style-extractor')
class StyleExtractor(StyleExtractorLayer, StyleMultiLayer):
	def __init__(self, A, **kwargs):

		merge_style = A.pull('merge-style', True)
		
		super().__init__(A, **kwargs)
		
		self.style_layers = [layer for layer in self.layers if isinstance(layer, StyleExtractorLayer)]
		self.merge_style = merge_style
		
		
	def collect_style(self):
		styles = []
		for layer in self.style_layers:
			styles.append(layer.collect_style())
			
		if self.merge_style:
			styles = torch.cat(styles, 1)
		return styles
	
	
	def extract(self, inp, **unused):
		
		content = super(StyleExtractorLayer, self).forward(inp)
		style = self.collect_style()
		
		return content, style







from .datasets import MNIST

from .model import load_model
# from .data import load_data
# from .data import load_data, Dataset
from .evaluation import evaluate
from .training import iterative_training
from .report import get_report
from .runs import Run
from .loading import Torch_Run
from . import extensions
from .report import get_report
from .analysis import Run_Manager
from . import records
from . import download





import sys, os
import time
from torch.utils.tensorboard import SummaryWriter
import tensorflow as tf
import scipy.misc
from io import BytesIO  # Python 3.x
from PIL import Image

import omnifig as fig

@fig.AutoComponent('logger')
class Logger(object):

	def __init__(self, log_dir=None, tensorboard=False, txt=False, logfile=None, **kwargs):

		self.tblog = None
		if tensorboard:
			try:
				self.tblog = SummaryWriter(log_dir=log_dir, **kwargs)
			except:
				self.tblog = None

		self.txtlog = None
		if txt:
			if logfile is None:
				logfile = open(os.path.join(log_dir, 'logfile.txt'), 'a+')
			self._old_stdout = sys.stdout
			self.txtlog = Tee(self._old_stdout, logfile)
			sys.stdout = self.txtlog

			now = time.strftime("%b-%d-%Y-%H%M%S")
			title = '**** Beginning Log {} ****\n'.format(now)
			title_stars = '*' *(len(title ) -1) + '\n'

			self.txtlog.write(title_stars + title + title_stars, logonly=True)

		self.global_step = None
		self.tag_fmt = None
		

	def set_step(self, step):
		self.global_step = step

	def get_step(self):
		return self.global_step

	def set_tag_format(self, fmt=None):
		self.tag_fmt = fmt
		
	def get_tag_format(self):
		return self.tag_fmt

	def add_hparams(self, param_dict, metrics={}):
		self.tblog.add_hparams(param_dict, metrics)

	def add(self, data_type, tag, *args, global_step=None, **kwargs): # TODO: test and maybe clean

		if self.tblog is None:
			return None

		# if data_type == 'scalar' and not isinstance(args[0], float):
		# 	print(tag, type(args[0]), args[0])
		# 	assert False

		add_fn = self.tblog.__getattribute__('add_{}'.format(data_type))

		if global_step is None:
			global_step = self.global_step

		if self.tag_fmt is not None:
			tag = self.tag_fmt.format(tag)

		add_fn(tag, *args, global_step=global_step, **kwargs)

	def flush(self):
		if self.tblog is not None:
			self.tblog.flush()
		if self.txtlog is not None:
			self.txtlog.flush()


### Write to stdout and log file
class Tee(object):
	def __init__(self, stdout, logfile):
		self.stdout = stdout
		self.logfile = logfile

	def write(self, obj, logonly=False):
		self.logfile.write(obj)
		if logonly:
			return
		self.stdout.write(obj)

	def flush(self):
		self.stdout.flush()

	def __del__(self):
		self.logfile.close()





class Old_Logger(object):
	def __init__(self, log_dir, tensorboard=True, txt=False, auto_step=False, step0=0, step_delta=1, logfile=None):
		raise Exception('Deprecated, use Logger instead')
		now = time.strftime("%b-%d-%Y-%H%M%S")

		self.tblog = TBLogger(log_dir) if tensorboard else None

		if txt:
			if logfile is None:
				logfile = open(os.path.join(log_dir, 'logfile.txt'), 'a+')
			self._old_stdout = sys.stdout
			self.txtlog = Tee(self._old_stdout, logfile)

			title = '**** Beginning Log {} ****\n'.format(now)
			title_stars = '* ' *(len(title ) -1) + '\n'

			self.txtlog.write(title_stars + title + title_stars, logonly=True)

		self.auto_step = auto_step
		self.step = step0
		self.delta = step_delta

	def update(self, info, step=None):
		if self.tblog is None:
			return

		if step is None:
			step = self.step
		self.step += self.delta

		if isinstance(info, StatsMeter):
			info = {name: info[name].val.item() for name in info.keys()}

		for k ,v in info.items():
			self.tblog.scalar_summary(k, v, step)

	def update_images(self, info, step=None): # input 1 x H x W x 3 numpy images
		if self.tblog is None:
			return

		if step is None:
			step = self.step

		for k ,v in info.items():
			self.tblog.image_summary(k, v, step)


class TBLogger(object):
	def __init__(self, log_dir):
		"""Create a summary writer logging to log_dir."""
		self.writer = tf.summary.FileWriter(log_dir +'/')
		self.tag_steps = {}

	def scalar_summary(self, tag, value, step=None):
		"""Log a scalar variable."""
		if step is None:
			if tag not in self.tag_steps:
				self.tag_steps[tag] = 0
			self.tag_steps[tag] += 1
			step = self.tag_steps[tag]
		summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
		self.writer.add_summary(summary, step)

	def image_summary(self, tag, images, step):
		"""Log a list of images."""

		img_summaries = []
		for i, img in enumerate(images):
			# Write the image to a string
			s = BytesIO()
			Image.fromarray(img).save(s, format="jpeg")

			# Create an Image object
			img_sum = tf.Summary.Image(encoded_image_string=s.getvalue(),
			                           height=img.shape[0],
			                           width=img.shape[1])
			# Create a Summary value
			img_summaries.append(tf.Summary.Value(tag='%s/%d' % (tag, i), image=img_sum))

		# Create and write Summary
		summary = tf.Summary(value=img_summaries)
		self.writer.add_summary(summary, step)

	def histo_summary(self, tag, values, step, bins=1000):
		"""Log a histogram of the tensor of values."""

		# Create a histogram using numpy
		counts, bin_edges = np.histogram(values, bins=bins)

		# Fill the fields of the histogram proto
		hist = tf.HistogramProto()
		hist.min_val = float(np.min_val(values))
		hist.max = float(np.max(values))
		hist.num = int(np.prod(values.shape))
		hist.sum = float(np.sum(values))
		hist.sum_squares = float(np.sum(values ** 2))

		# Drop the start of the first bin
		bin_edges = bin_edges[1:]

		# Add bin edges and counts
		for edge in bin_edges:
			hist.bucket_limit.append(edge)
		for c in counts:
			hist.bucket.append(c)

		# Create and write Summary
		summary = tf.Summary(value=[tf.Summary.Value(tag=tag, histo=hist)])
		self.writer.add_summary(summary, step)
		self.writer.flush()






import sys, os

from contextlib import nullcontext, redirect_stdout, redirect_stderr


import omnifig as fig

# class LogFile_Arg(fig.Meta_Argument, name='logfile', code='l', num_params=1):
# 	pass
#
# class Logging_Mode(fig.Run_Mode, supports=[LogFile_Arg]):
#
# 	# def __init__(self, meta, config, auto_meta_args=[]):
# 	# 	pass
#
# 	def run(self, script_info, meta, config):
#
# 		ctxt, ctxt2 = nullcontext(), nullcontext()
#
# 		if 'JOBDIR' in os.environ:
# 			proc = os.environ['PROCESS_ID']
# 			logpath = os.path.join(os.environ['JOBDIR'], f'out-{proc}.txt')
#
# 			ctxt = redirect_stderr(sys.stdout)
# 			ctxt2 = redirect_stdout(open(logpath, 'a+'))
#
# 		with ctxt, ctxt2:
# 			return super().run(script_info, meta, config)
#
# 	pass







import torch
from collections import OrderedDict

from .features import DeviceBase, Statelike, Deviced

def to(obj, device):
	try:
		return obj.to(device)
	except AttributeError:
		if isinstance(obj, (list, tuple)):
			return TensorList([to(o, device) for o in obj])
		if isinstance(obj, dict):
			return TensorDict({k:to(v,device) for k,v in obj.items()})
		raise Exception('Unknown object {}: {}'.format(type(obj), obj))

class Movable(DeviceBase):
	def size(self, *args, **kwargs):
		raise NotImplementedError

class TensorDict(Movable, OrderedDict):

	def __init__(self, *args, _size_key=None, **kwargs):
		super().__init__(*args, **kwargs)
		self.__dict__['_size_key'] = _size_key

	def to(self, device):
		super().to(device)
		for k,v in self.items():
			try:
				self[k] = v.to(device)
			except AttributeError:
				pass

		return self

	def detach(self):
		for k,v in self.items():
			try:
				self[k] = v.detach()
			except AttributeError:
				pass

		return self

	def _find_size_key(self):
		for k, v in self.items():
			if isinstance(v, torch.Tensor):
				self._size_key = k
				return

		raise Exception('No torch.Tensor found')

	def size(self, *args, **kwargs):
		if self._size_key is None:
			self._find_size_key()
		return self[self._size_key].size(*args, **kwargs)

	def __getstate__(self):
		return {k:v for k,v in self.items()}

	def __setstate__(self, state):
		self.update(state)

	def __getattr__(self, item):
		if item in self.__dict__ or item in {'__setattr__', '__setstate__'}:
			return super().__getattribute__(item)
		return self.__getitem__(item)

	def __setattr__(self, key, value):
		if key in self.__dict__:
			return super().__setattr__(key, value)
		return self.__setitem__(key, value)

	def __delattr__(self, item):
		if item in self.__dict__:
			return super().__delattr__(item)
		return self.__delitem__(item)

class TensorList(Movable, list):

	def __init__(self, *args, _size_key=None, **kwargs):
		super().__init__(*args, **kwargs)
		self._size_key = _size_key

	def to(self, device):
		super().to(device)
		for i,x in enumerate(self):
			try:
				self[i] = x.to(device)
			except AttributeError:
				pass

		return self

	def _find_size_key(self):
		for i, x in enumerate(self):
			if isinstance(x, torch.Tensor):
				self._size_key = i
				return

		raise Exception('No torch.Tensor found')

	def size(self, *args, **kwargs):
		if self._size_key is None:
			self._find_size_key()
		return self[self._size_key].size(*args, **kwargs)

	def __getstate__(self):
		return [x for x in self]

	def __setstate__(self, state):
		for i,x in enumerate(state):
			self[i] = x

class Cached(Deviced):
	def __init__(self, A, **kwargs):
		self.volatile = TensorDict()
		super().__init__(A, **kwargs)
	
	def to(self, device):
		super().to(device)
		self.volatile.to(self.get_device())


class TrackedAttrs(Statelike, Cached):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		
		self._saveable_attrs = set()
	
	def register_attr(self, name, data=None, update=True):
		self._saveable_attrs.add(name)
		if update:
			setattr(self, name, data)
	
	def state_dict(self, *args, **kwargs): # TODO: prevent overlapping parameters collection
		volatile = self.volatile
		self.volatile = None
		
		try:
			data = super().state_dict(*args, **kwargs)
		except AttributeError:
			data = None
		
		attrs = {}
		for name in self._saveable_attrs:
			try:
				val = getattr(self, name, None)
			except AttributeError:
				pass
			else:
				attrs[name] = val.state_dict() if isinstance(val, Statelike) else val
		
		self.volatile = volatile
		return {'parameters': data, 'attrs': attrs}
	
	def load_state_dict(self, state_dict, strict=True):
		missing_attrs = []
		for name, data in state_dict.get('attrs', {}).items():
			try:
				val = getattr(self, name)
			except AttributeError:
				if data is not None:
					missing_attrs.append(name)
				else:
					setattr(self, name, None)
			else:
				if isinstance(val, Statelike):
					val.load_state_dict(data)
				else:
					setattr(self, name, data)
		if strict and len(missing_attrs):
			raise RuntimeError('Missing registered attributes: {}'.format(', '.join(missing_attrs)))
		try:
			return super().load_state_dict(state_dict['parameters']
			                               if 'parameters' in state_dict
			                               else state_dict, strict=strict)
		except AttributeError:
			raise
	
	def __setattr__(self, key, value):
		if isinstance(value, Statelike):
			self.register_attr(key, value, update=False)
		super().__setattr__(key, value)
	
	def __delattr__(self, item):
		if item in self._saveable_attrs:
			self._saveable_attrs.discard(item)
		return super().__delattr__(item)







from pathlib import Path
import torch

import omnifig as fig

from .. import util
from .runs import Run
# from .data impor

_config_root = Path(__file__).parents[2] / 'config'


# @fig.AutoModifier('torch')
@fig.Component('run')
class Torch_Run(Run):
	
	def has_results(self, ident, path=None, ext=None):
		if ext is None:
			ext = 'pth.tar'
		return super().has_results(ident, path=path, ext=ext)
	
	def _save_results(self, data, path=None, name=None, ext='pth.tar'):
		if path is None:
			path = self.get_path()
		
		path = Path(path)
		if path.is_dir():
			assert name is not None, 'name is missing'
			path = path / f'{name}.{ext}'
		
		return torch.save(data, str(path))
	
	def _load_results(self, path=None, name=None, ext='pth.tar', device=None, **kwargs):
		if path is None:
			path = self.get_path()
		
		path = Path(path)
		if path.is_dir():
			assert name is not None, 'name is missing'
			path = path / f'{name}.{ext}'
			
		
		special = {'map_location':device} if device is not None else {}
		return torch.load(str(path), **special)


def respect_config(A):
	device = A.push('device', 'cuda' if torch.cuda.is_available() else 'cpu',
	                overwrite=not torch.cuda.is_available())
	
	cudnn_det = A.pull('cudnn_det', False)
	if 'cuda' in device:
		torch.backends.cudnn.deterministic = cudnn_det
	
	A.push('seed', util.gen_random_seed(), overwrite=False, silent=False)
	
#
# @fig.Script('load_config')
# def load_config(A):
#
#
# 	# override = A.override if 'override' in A else {}
# 	override = A.pull('override', {})
#
# 	extend = A.pull('extend', None)
#
# 	raw_path, loading = get_raw_path(A)
# 	if raw_path is None:
# 		return A
#
# 	print(f'Loading Config: {raw_path}')
#
# 	path = find_config(raw_path)
#
# 	load_A = fig.get_config(path)
# 	if loading:
# 		load_A.update(A)
# 	else:
# 		A = load_A
# 	A.update(override)
#
# 	A.push('load' if loading else 'resume', raw_path, overwrite=True)
#
# 	if extend is not None:
# 		A.push('training.step_limit', extend)
#
# 	return A

#
# @fig.Script('load_records')
# def load_records(A):
#
# 	last = A.pull('last', False)
#
# 	path = A.pull('path', None)
# 	loading = False
#
# 	if path is None:
# 		path, loading = get_raw_path(A)
# 	if path is None or loading:
# 		return setup_records(A)
#
# 	print(f'Loading Records: {path}')
#
# 	ckpt_path = find_checkpoint(path, last=last)
# 	A.push('model._load_params', ckpt_path, overwrite=True)
#
# 	path = find_records(path, last=last)
#
# 	with open(path, 'r') as f:
# 		records = yaml.safe_load(f)
# 	return records
#

# fig.register_config_dir(str(_config_root)) # WARNING: overrides debug
fig.register_config('origin', str(_config_root/'origin.yaml'))





import os

import h5py as hf

from omnibelt import unspecified_argument

import numpy as np
import torch
from torch.nn import functional as F
import torchvision

from ... import util
from ...data import Dataset, Batchable, Deviced, Downloadable, Image_Dataset


from .transforms import Interpolated





class Torchvision_Toy_Dataset(Batchable, Deviced, Downloadable, Image_Dataset): # TODO: enable (pytorch) transforms
	available_modes = {'train', 'test'}

	_default_label_attr = 'targets'

	def __init__(self, A, dataroot=None, mode=None, labeled=None, label_attr=unspecified_argument,
	             din=None, dout=None,
	             dataset_kwargs=None, _req_kwargs=None, **unused):
		'''
		Requires dataset object to wrap it (since this is a `util.Simple_Child`).

		Images must be in `dataset.data` and labels in `dataset.[label_attr]` (if provided).

		:param dataset: compatible pytorch torchvision dataset object
		:param train: bool
		:param label_attr: attr name used to access labels in `dataset`
		:param din: optional (if it has to be overwritten)
		:param dout: optional (if it has to be overwritten)
		:param root: path to dataset files
		'''
		
		if dataroot is None:
			dataroot = self._get_dataroot(A)
			
		if mode is None:
			mode = A.pull('mode', 'train')
		
		if dataset_kwargs is None:
			dataset_kwargs = self._get_dataset_kwargs(A, mode=mode, **unused)
		if 'root' not in dataset_kwargs:
			dataset_kwargs['root'] = str(dataroot)

		resize = A.pull('resize', True)
		C, H, W = self.din
		if resize and (H != 32 or W != 32):
			self.din = C, 32, 32

		if label_attr is unspecified_argument:
			if labeled is None:
				labeled = A.pull('labeled', True)
			label_attr = A.pull('label_attr', self._default_label_attr) if labeled else None

		if label_attr is None:
			dout = self.din if din is None else din

		if _req_kwargs is None:
			_req_kwargs = dataset_kwargs
		else:
			_req_kwargs.update(dataset_kwargs)

		super().__init__(A, din=din, dout=dout, dataroot=dataroot,
		                 _req_kwargs=_req_kwargs,
		                 **unused)
		self.add_existing_modes('test')

		self.labeled = label_attr is not None

		# self.root = root

		images = self.data
		if isinstance(images, np.ndarray):
			images = torch.from_numpy(images)
		images = images.float().div(255)
		if images.ndimension() == 3:
			images = images.unsqueeze(1)
		
		if resize:
			images = F.interpolate(images, (32, 32), mode='bilinear')
		
		if label_attr is None:
			self.labels = None
		else:
			labels = getattr(self, label_attr)
			if not isinstance(labels, torch.Tensor):
				labels = torch.tensor(labels)
			self.register_buffer('labels', labels)
			# delattr(self, label_attr)

		self.register_buffer('images', images)
		# del self.data

	def download(cls, A, **kwargs):
		cls(A, download=True)

	def get_raw_data(self):
		if self.labeled:
			return self.images, self.labels
		return self.images

	def __len__(self):
		return len(self.images)

	def __getitem__(self, item):
		img = self.images[item]
		if self.labeled:
			return img, self.labels[item]
		return img
	
	@classmethod
	def _get_dataroot(cls, A=None, ident=None, silent=False):
		dataroot = util.get_data_dir(A, silent=silent)
		if ident is None:
			assert hasattr(cls, '_dataset_ident'), f'{cls.__name__} must be registered if no ident is provided'
			ident = cls._dataset_ident
		return dataroot / ident
	
	@classmethod
	def _get_dataset_kwargs(cls, A, mode=None, download=None, train=None, **unused):
		
		if mode is None:
			mode = A.pull('mode', 'train')
		
		if download is None:
			download = A.pull('download', False)
		
		if train is None:
			train = A.pull('train', mode != 'test')
		
		return {'download': download, 'train':train}



@Dataset('mnist')
class MNIST(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.MNIST):
	din = (1, 28, 28)
	dout = 10

@Dataset('kmnist')
class KMNIST(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.KMNIST):
	din = (1, 28, 28)
	dout = 10

@Dataset('fmnist')
class FashionMNIST(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.FashionMNIST):
	din = (1, 28, 28)
	dout = 10

@Dataset('emnist')
class EMNIST(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.EMNIST):
	din = (1, 28, 28)
	dout = 26
	
	_split_keys = {
		'byclass': '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxy',
		'bymerge': '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabdefghnqrt',
		'letters': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
		'digits': '0123456789',
		'mnist': '0123456789',
		'balanced': '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabdefghnqrt',
	}
	
	def __init__(self, A, **kwargs):
		
		# split = A.pull('group', 'letters')
		# if split != 'letters':
		# 	raise NotImplementedError
		
		# dataset = torchvision.datasets.EMNIST(str(dataroot), split=split, **kwargs)


		# labeled = A.pull('labeled', False)
		
		
		# labels_key = self._split_keys.get(split, None) if labeled else None
		# dout = len(labels_key) if labeled else None
		
		# TODO: Automodifier for selected_classes
		# selected_classes = A.pull('selected_classes', None)
		# if selected_classes is not None:
		# 	sel = None
		# 	lbls = dataset.targets.clone()
		# 	full_key = labels_key
		# 	labels_key = []
		#
		# 	for i, c in enumerate(selected_classes):
		#
		# 		s = lbls == c
		# 		if sel is None:
		# 			sel = s
		# 		else:
		# 			sel += s
		# 		dataset.targets[s] = i
		# 		if full_key is not None:
		# 			labels_key.append(full_key[c])
		#
		# 	dataset.targets = dataset.targets[sel]
		# 	dataset.data = dataset.data[sel]
		#
		# 	if labeled:
		# 		dout = len(selected_classes)

		super().__init__(A, **kwargs)
		
		self.images = self.images.permute(0,1,3,2)
		
		if self.split == 'letters' and self.labels is not None:
			self.labels -= 1  # targets use 1-based indexing :(
		
		self.labels_key = self._split_keys[self.split]
	
	@classmethod
	def _get_dataset_kwargs(cls, A, split=None, **other):
		kwargs = super()._get_dataset_kwargs(A, **other)
		if split is None:
			split = A.pull('group', 'letters')
		kwargs['split'] = split
		return kwargs


@Dataset('svhn')
class SVHN(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.SVHN):
	din = (3, 32, 32)
	dout = 10
	
	_default_label_attr = 'labels'

	@classmethod
	def _get_dataset_kwargs(cls, A, **unused):
		kwargs = super()._get_dataset_kwargs(A, **unused)
		kwargs['split'] = 'train' if kwargs['train'] else 'test'
		del kwargs['train']
		return kwargs


@Dataset('cifar10')
class CIFAR10(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.CIFAR10):
	din = (3, 32, 32)
	dout = 10

	def _get_dataroot(cls, *args, ident=None, **kwargs):
		return super()._get_dataroot(*args, ident='cifar', **kwargs)


@Dataset('cifar100')
class CIFAR100(Torchvision_Toy_Dataset, util.InitWall, torchvision.datasets.CIFAR100):
	din = (3, 32, 32)
	dout = 100

	def _get_dataroot(cls, *args, ident=None, **kwargs):
		return super()._get_dataroot(*args, ident='cifar', **kwargs)




import sys, os
from pathlib import Path
import numpy as np
import torch
import copy
import torch.nn as nn

from omnibelt import primitives, InitWall

import omnifig as fig

from .. import util
from ..util import Configurable, Seed, Checkpointable, Switchable, TrackedAttrs, \
	Dimensions, Deviced, DeviceBase, DimensionBase
import torch.multiprocessing as mp
from itertools import chain

from .clock import AlertBase


class FunctionBase(DimensionBase, DeviceBase, InitWall, nn.Module):  # any differentiable vector function
	def __init__(self, din=None, dout=None, device=None, **unused):
		super().__init__(din=din, dout=dout, device=device, **unused)
	

class Function(Switchable, TrackedAttrs, Dimensions, Deviced, Configurable, FunctionBase):
	
	def switch_to(self, mode):
		super().switch_to(mode)
		if mode == 'train':
			self.train()
		else:
			self.eval()
	
	def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
		
		super()._load_from_state_dict(state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs)
	
		if len(unexpected_keys): # because pytorch doesn't like persistent buffers that are None
			persistent_buffers = {k: v for k, v in self._buffers.items() if k not in self._non_persistent_buffers_set}
			for name, val in persistent_buffers.items():
				if val is None:
					key = prefix + name
					if key in unexpected_keys:
						setattr(self, name, state_dict[key])
						unexpected_keys.remove(key)
				
	
	def get_hparams(self):
		return {}

class HyperParam(Function):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._hparams = set()
	
	def register_hparam(self, name, val):
		# if not isinstance(val, util.ValueBase):
		# 	val = util.ValueBase(val)
		self._hparams.add(name)
		self.register_attr(name, val)
		
	def state_dict(self, *args, **kwargs):
		state = super().state_dict(*args, **kwargs)
		state['hparams'] = list(self._hparams)
		return state
		
	def load_state_dict(self, state_dict, strict=True):
		if strict:
			assert len(self._hparams) == len(state_dict['hparams']) \
			       == len(self._hparams.intersection(set(state_dict['hparams'])))
		return super().load_state_dict(state_dict, strict=strict)
	
	def get_hparams(self):
		hparams = {}
		for name in self._hparams:
			try:
				val = getattr(self, name, None)
			except AttributeError:
				continue
			hparams[name] = val if type(val) in primitives else val.item()
		return hparams


class Savable(Checkpointable, Function):
	
	def __init__(self, A, **kwargs):
		strict_load_state = A.pull('strict_load_state', True)
		
		super().__init__(A, **kwargs)
		
		self._strict_load_state = strict_load_state
	
	def checkpoint(self, path, ident='model'):
		data = {
			'model_str': str(self),
			'model_state': self.state_dict(),
		}
		path = Path(path)
		if path.is_dir():
			path = path / f'{ident}.pth.tar'
		torch.save(data, str(path))
		
	def load_checkpoint(self, path, ident='model', _data=None):
		if _data is None:
			path = Path(path)
			if path.is_dir():
				path = path / f'{ident}.pth.tar'
			data = torch.load(str(path), map_location=self.get_device())
		else:
			data = _data
		self.load_state_dict(data['model_state'], strict=self._strict_load_state)
		return path
	

		
class Recordable(util.StatsClient, Function):
	pass
	
class Maintained(Function, AlertBase):
	
	def maintain(self, tick, info=None):
		pass
	
	def activate(self, tick, info=None):
		return self.maintain(tick, info=info)

class Visualizable(Recordable):
	def visualize(self, info, logger): # records output directly to logger
		with torch.no_grad():
			self._visualize(info, logger)
			
	def _visualize(self, info, logger):
		pass # by default nothing is visualized


class Evaluatable(Recordable): # TODO: maybe not needed

	def evaluate(self, info, config=None, out=None):
		if config is None:
			config = info.get_config()
		return self._evaluate(info, config, out=out)

	def _evaluate(self, info, config, out=None):
		if out is None:
			out = util.TensorDict()
		return out # by default eval does nothing


@fig.AutoModifier('optim')
class Optimizable(Function):

	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		self.optim = None

	def set_optim(self, A=None):
		
		if self.optim is not None:
			return
		
		sub_optims = {}
		for name, child in self.named_children():
			if isinstance(child, Optimizable) and child.optim is not None:
				sub_optims[name] = child.optim

		if len(sub_optims):
			params = [] # param eters not already covered by sub_optims
			for name, param in self.named_parameters():
				name = name.split('.')[0]
				if name not in sub_optims:
					params.append(param)
		
			if len(sub_optims) == 1 and len(params) == 0: # everything convered by a single sub_optim
				optim = next(iter(sub_optims.values()))
			else:
				if len(params):
					if 'me' in sub_optims:
						raise Exception('invalid names: {} already in {}'.format('me', sub_optims.keys()))
					# sub_optims['me'] = util.default_create_optim(params, optim_info)
					sub_optims['me'] = A.pull('optim')
					sub_optims['me'].prep(params)
				optim = util.Complex_Optimizer(**sub_optims)
		else:
			optim = A.pull('optim')
			optim.prep(self.parameters())
			# optim = util.default_create_optim(self.parameters(), optim_info)
			
		self.optim = optim
		
		return optim

	def _optim_step(self, loss): # should only be called during training
		if self.optim is None:
			raise Exception('Optimizer not set')
		self.optim.zero_grad()
		loss.backward()
		self.optim.step()

	def load_state_dict(self, state_dict, strict=True):
		super().load_state_dict(state_dict['model'], strict=strict)
		if self.optim is not None and 'optim' in state_dict:
			self.optim.load_state_dict(state_dict['optim'])

	def state_dict(self, *args, **kwargs):
		state_dict = {
			'model': super().state_dict(*args, **kwargs),
		}
		if self.optim is not None:
			state_dict['optim'] = self.optim.state_dict()
		return state_dict


class Trainable(Maintained, HyperParam, Recordable, Optimizable, Function, AlertBase):
	
	def __init__(self, A, **kwargs):
		
		optim_metric = A.pull('optim-metric', 'loss')
		
		super().__init__(A, **kwargs)
		
		self.optim_metric = optim_metric
		self.register_stats(self.optim_metric)
	
	def activate(self, tick, info=None):
		batch = info.get_batch()
		info.receive_output(batch, self.step(batch))
		return super().activate(tick, info=info)
	
	def step(self, batch, **kwargs):  # Override pre-processing mixins
		out = self._step(batch, **kwargs)
		# if self.train_me() and self.optim_metric in out:
		if self.optim_metric in out:
			self.mete(self.optim_metric, out[self.optim_metric])
		return out

	def _step(self, batch, out=None):  # Override post-processing mixins
		if out is None:
			out = util.TensorDict()
		return out

	def get_metric(self):
		if self.optim_metric is not None:
			return self.get_stat(self.optim_metric)

	def get_description(self):
		progress = ''
		if self.optim_metric is not None:
			metric = self.get_metric()
			if metric is not None and metric.count > 0:
				progress = f'{self.optim_metric}: {metric.val:.3f} ({metric.smooth:.3f})'
				
		return progress

	# NOTE: never call an optimizer outside of _step (not in mixinable functions)
	# NOTE: before any call to an optimizer check with self.train_me()
	def train_me(self):
		return self.training and self.optim is not None
	

class Model(Seed, Savable, Trainable, Evaluatable, Visualizable, Function): # top level - must be implemented to train
	pass




class Regularizable(object):
	def regularize(self, q):
		return torch.tensor(0).type_as(q)



@fig.AutoModifier('generative')
class Generative(object):
	def sample_prior(self, N=1):
		raise NotImplementedError

	def generate(self, N=1, prior=None):
		if prior is None:
			prior = self.sample_prior(N)
		return self(prior)


@fig.AutoModifier('encodable')
class Encodable(object):
	def encode(self, x): # by default this is just forward pass
		return self(x)


@fig.AutoModifier('decodable')
class Decodable(object): # by default this is just the forward pass
	def decode(self, q):
		return self(q)


@fig.AutoModifier('invertible')
class Invertible(object):
	def inverse(self, *args, **kwargs):
		raise NotImplementedError



